/*
 * ConversionHandler.cpp
 *
 *  Created on: 23. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "ConversionHandler.h"

#include "automaton/convert/ToRegExpStateElimination.h"
#include "automaton/convert/ToRegExpAlgebraic.h"
#include "automaton/convert/ToGrammarLeftRG.h"
#include "automaton/convert/ToGrammarRightRG.h"

#include "regexp/convert/ToAutomatonGlushkov.h"
#include "regexp/convert/ToAutomatonThompson.h"
#include "regexp/convert/ToAutomatonDerivation.h"

#include "grammar/convert/ToAutomaton.h"

#include "grammar/convert/ToRegExpAlgebraic.h"

#include "regexp/convert/ToGrammarRightRGGlushkov.h"
#include "regexp/convert/ToGrammarRightRGDerivation.h"

#include <factory/XmlDataFactory.hpp>

ConversionHandler::ConversionHandler( std::list<sax::Token> & tokens, const std::string & target, const std::string & algorithm ) :
	m_tokens( tokens ),
	m_source( parseFormalismFromTokens( ) ),
	m_target( parseFormalismFromString( target ) ),
	m_algorithm( parseAlgorithmFromString( algorithm ) )
{

}

void ConversionHandler::convert( void )
{
	if( m_source == FINITE_AUTOMATON )
		convertFA( );
	else if( m_source == REGULAR_EXPRESSION )
		convertRE( );
	else if( m_source == REGULAR_GRAMMAR )
		convertRG( );
	else
		throw exception::AlibException( "ConversionHandler: Unknown source formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFA( void )
{
	if( m_target == REGULAR_EXPRESSION )
		convertFAtoRE( );
	else if( m_target == REGULAR_GRAMMAR )
		convertFAtoRG( );
	else
		throw exception::AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRE( void )
{
	if( m_target == FINITE_AUTOMATON )
		convertREtoFA( );
	else if( m_target == REGULAR_GRAMMAR )
		convertREtoRG( );
	else
		throw exception::AlibException( "ConversionHandler: Invalid target formalism." );
}

void ConversionHandler::convertRG( void )
{
	if( m_target == FINITE_AUTOMATON )
		convertRGtoFA( );
	else if( m_target == REGULAR_EXPRESSION )
		convertRGtoRE( );
	else
		throw exception::AlibException( "ConversionHandler: Invalid target formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFAtoRE( void )
{
	const automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(m_tokens);

	switch( m_algorithm ) {
	case BRZOZOWSKI_ALGEBRAIC:
		alib::XmlDataFactory::toStdout(automaton::convert::ToRegExpAlgebraic::convert( automaton ) );
		break;
	case STATE_ELIMINATION:
	case DEFAULT:
		alib::XmlDataFactory::toStdout(automaton::convert::ToRegExpStateElimination::convert(automaton));
		break;
	default:
		throw exception::AlibException( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertFAtoRG( void )
{
	const automaton::Automaton fsm = alib::XmlDataFactory::fromTokens<automaton::Automaton>( m_tokens );

	switch( m_algorithm ) {
	case INCOMING_TRANSITIONS:
		alib::XmlDataFactory::toStdout(automaton::convert::ToGrammarLeftRG::convert(fsm));
		break;
	case OUTGOING_TRANSITIONS:
	case DEFAULT:
		alib::XmlDataFactory::toStdout(automaton::convert::ToGrammarRightRG::convert(fsm));
		break;
	default:
		throw exception::AlibException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertREtoFA( void )
{
	const regexp::RegExp regexp = alib::XmlDataFactory::fromTokens<regexp::RegExp>(m_tokens);

	switch( m_algorithm ) {
	case BRZOZOWSKI_DERIVATION:
		alib::XmlDataFactory::toStdout(regexp::convert::ToAutomatonDerivation::convert(regexp));
		break;
	case THOMPSON_NFA:
		alib::XmlDataFactory::toStdout(regexp::convert::ToAutomatonThompson::convert(regexp));
		break;
	case GLUSHKOV_NFA:
	case DEFAULT:
		alib::XmlDataFactory::toStdout(regexp::convert::ToAutomatonGlushkov::convert(regexp));
		break;
	default:
		throw exception::AlibException( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertREtoRG( void )
{
	const regexp::RegExp regexp = alib::XmlDataFactory::fromTokens<regexp::RegExp>(m_tokens);

	switch(m_algorithm) {
	case BRZOZOWSKI_DERIVATION:
		alib::XmlDataFactory::toStdout(regexp::convert::ToGrammarRightRGDerivation::convert(regexp));
		break;
	case GLUSHKOV_NFA:
	case DEFAULT:
		alib::XmlDataFactory::toStdout(regexp::convert::ToGrammarRightRGGlushkov::convert(regexp));
		break;
	default:
		throw exception::AlibException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertRGtoFA( void ) {
	const grammar::Grammar grammar = alib::XmlDataFactory::fromTokens<grammar::Grammar>(m_tokens);
	switch( m_algorithm ) {
	case DEFAULT:
		alib::XmlDataFactory::toStdout(grammar::convert::ToAutomaton::convert(grammar));
		break;
	default:
		throw exception::AlibException( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertRGtoRE( void ) {
	const grammar::Grammar grammar = alib::XmlDataFactory::fromTokens<grammar::Grammar>(m_tokens);

	switch( m_algorithm ) {
	case BRZOZOWSKI_ALGEBRAIC:
	case DEFAULT:
		alib::XmlDataFactory::toStdout(grammar::convert::ToRegExpAlgebraic::convert(grammar));
	       break;
	default:
		throw exception::AlibException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromString( const std::string & target ) const {
	if( target == "fa" ) return FINITE_AUTOMATON;
	if( target == "re" ) return REGULAR_EXPRESSION;
	if( target == "rg" ) return REGULAR_GRAMMAR;

	throw exception::AlibException( "ConversionHandler: Unknown target type." );
}

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromTokens( void ) const {
	if( alib::XmlDataFactory::first<automaton::Automaton>(m_tokens) )
		return FINITE_AUTOMATON;

	if( alib::XmlDataFactory::first<regexp::RegExp>(m_tokens) )
	        return REGULAR_EXPRESSION;

	if( alib::XmlDataFactory::first<grammar::Grammar>(m_tokens) )
		return REGULAR_GRAMMAR;

	throw exception::AlibException( "ConversionHandler: Invalid input formalism." );
}

ConversionHandler::TAlgorithm ConversionHandler::parseAlgorithmFromString( const std::string & algorithm ) const {
	if( algorithm == "elimination" ) return STATE_ELIMINATION;
	if( algorithm == "algebraic" ) return BRZOZOWSKI_ALGEBRAIC;
	if( algorithm == "brzozowski" ) return BRZOZOWSKI_DERIVATION;
	if( algorithm == "glushkov" ) return GLUSHKOV_NFA;
	if( algorithm == "thompson" ) return THOMPSON_NFA;
	if( algorithm == "incoming" ) return INCOMING_TRANSITIONS;
	if( algorithm == "outgoing" ) return OUTGOING_TRANSITIONS;

	if( algorithm == "" || algorithm == "default" ) return DEFAULT;

	throw exception::AlibException( "ConversionHandler:: Invalid algorithm. See help." );
}
