/*
 * ToAutomaton.cpp
 *
 * Created on: 1. 11. 2013
 * Author: Tomas Pecka
 */
#include "ToAutomaton.h"

#include <alphabet/LabeledSymbol.h>
#include <exception/AlibException.h>

namespace grammar {

namespace convert {

automaton::Automaton ToAutomaton::convert(const grammar::Grammar& grammar)
{
	 automaton::Automaton* out = NULL;
	 grammar.getData().Accept((void*) &out, ToAutomaton::TO_AUTOMATON);
	 automaton::Automaton res = std::move(*out);
	 delete out;
	 return res;
}

automaton::NFA ToAutomaton::convert(const grammar::LeftRG& grammar)
{
	std::map<alphabet::Symbol, automaton::State> stateMap;
	std::set<automaton::State> states;

	// step 2
	for(const auto& symbol : grammar.getNonterminalAlphabet())
	{
		automaton::State state(static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel());
		states.insert(state);
		stateMap.insert(std::make_pair(symbol, state));
	}

	// step 1, 4
	const automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	states.insert(q0);
	automaton::NFA automaton(q0);
	automaton.setInputAlphabet(grammar.getTerminalAlphabet());
	automaton.setStates(states);

	// step 3
	for(const auto& rule : grammar.getRules())
	{
		const alphabet::Symbol& lhs = rule.first;
		for(const auto& ruleRHS : rule.second)
		{
			if(ruleRHS.is<std::pair<alphabet::Symbol, alphabet::Symbol>>()) // if B->Ca => \delta(C,a)=B
			{
				const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = ruleRHS.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
				automaton.addTransition(stateMap.find(rhs.first)->second, rhs.second, stateMap.find(lhs)->second);
			}
			else // if B->a => \delta(StartState,a)=B
			{
				const alphabet::Symbol& rhs = ruleRHS.get<alphabet::Symbol>();
				automaton.addTransition(q0, rhs, stateMap.find(lhs)->second);
			}
		}
	}

	// step 5
	automaton.addFinalState(stateMap.find(grammar.getInitialSymbol())->second);
	if(grammar.getGeneratesEpsilon())
		automaton.addFinalState(q0);

	return automaton;
}

automaton::NFA ToAutomaton::convert(const grammar::RightRG& grammar)
{
	std::map<alphabet::Symbol, automaton::State> stateMap;
	std::set<automaton::State> states;

	// step2
	for(const auto& symbol : grammar.getNonterminalAlphabet())
	{
		automaton::State state(static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel());
		states.insert(state);
		stateMap.insert(std::make_pair(symbol, state));
	}

	// step 1, 4
	const automaton::State AState = automaton::createUniqueState(automaton::State("A"), states);
	states.insert(AState);
	automaton::NFA automaton(stateMap.find(grammar.getInitialSymbol())->second);
	automaton.setStates(states);

	automaton.setInputAlphabet(grammar.getTerminalAlphabet());

	// step 3
	for(const auto& rule : grammar.getRules())
	{
		const alphabet::Symbol& lhs = rule.first;
		for(const auto& ruleRHS : rule.second)
		{
			if(ruleRHS.is<std::pair<alphabet::Symbol, alphabet::Symbol>>()) // if B->aC => \delta(B,a)=C
			{
				const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = ruleRHS.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
				automaton.addTransition(stateMap.find(lhs)->second, rhs.first, stateMap.find(rhs.second)->second);
			}
			else // if B->a => \delta(B,a)=AState
			{
				const alphabet::Symbol& rhs = ruleRHS.get<alphabet::Symbol>();
				automaton.addTransition(stateMap.find(lhs)->second, rhs, AState);
			}
		}
	}

	// step 5
	automaton.addFinalState(AState);
	if(grammar.getGeneratesEpsilon())
		automaton.addFinalState(automaton.getInitialState());

	return automaton;
}

void ToAutomaton::Visit(void* userData, const grammar::RightRG& grammar) const
{
	automaton::Automaton* & out = *((automaton::Automaton**) userData);
	out = new automaton::Automaton(this->convert(grammar));
}

void ToAutomaton::Visit(void* userData, const grammar::LeftRG& grammar) const
{
	automaton::Automaton* & out = *((automaton::Automaton**) userData);
	out = new automaton::Automaton(this->convert(grammar));
}

void ToAutomaton::Visit(void*, const grammar::RightLG&) const
{
	throw exception::AlibException("Unsupported grammar type RightLG");
}

void ToAutomaton::Visit(void*, const grammar::LeftLG&) const
{
	throw exception::AlibException("Unsupported grammar type LeftLG");
}

void ToAutomaton::Visit(void*, const grammar::LG& ) const {
	throw exception::AlibException("Unsupported grammar type LG");
}

void ToAutomaton::Visit(void*, const grammar::CFG& ) const {
	throw exception::AlibException("Unsupported grammar type CFG");
}

void ToAutomaton::Visit(void*, const grammar::EpsilonFreeCFG& ) const {
	throw exception::AlibException("Unsupported grammar type EpsilonFreeCFG");
}

void ToAutomaton::Visit(void*, const grammar::CNF& ) const {
	throw exception::AlibException("Unsupported grammar type CFG");
}

void ToAutomaton::Visit(void*, const grammar::GNF& ) const {
	throw exception::AlibException("Unsupported grammar type GNF");
}

void ToAutomaton::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void ToAutomaton::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void ToAutomaton::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void ToAutomaton::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const ToAutomaton ToAutomaton::TO_AUTOMATON;

} /* namespace convert */

} /* namespace grammar */
