/*
 * AutomatonToStringComposer.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "AutomatonToStringComposer.h"
#include "../exception/AlibException.h"

#include "../StringApi.hpp"

namespace automaton {

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const DFA& automaton, const State& from) const {
	std::map<std::pair<State, alphabet::Symbol>, State> symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, State>::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end()) {
			out << " -";
		} else {
			out << " ";
			alib::stringApi<label::Label>::compose(out, toStates->second.getName());
		}
	}
}

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const NFA& automaton, const State& from) const {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::stringApi<label::Label>::compose(out, to.getName());
				sign = true;
			}
		}
	}
}

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const MultiInitialStateNFA& automaton, const State& from) const {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::stringApi<label::Label>::compose(out, to.getName());
				sign = true;
			}
		}
	}
}

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const EpsilonNFA& automaton, const State& from) const {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getSymbolTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::stringApi<label::Label>::compose(out, to.getName());
				sign = true;
			}
		}
	}

	std::map<State, std::set<State> > epsilonTransitionsFromState = automaton.getEpsilonTransitionsFromState(from);
	if(epsilonTransitionsFromState.size() == 0 || epsilonTransitionsFromState.begin()->second.size() == 0) {
		out << " -";
	} else {
		bool sign = false;
		for(const State& to : epsilonTransitionsFromState.begin()->second) {
			out << (sign ? "|" : " ");
			alib::stringApi<label::Label>::compose(out, to.getName());
			sign = true;
		}
	}
}

void AutomatonToStringComposer::compose(std::ostream& out, const DFA& automaton) const {
	out << "DFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

void AutomatonToStringComposer::compose(std::ostream& out, const NFA& automaton) const {
	out << "NFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

void AutomatonToStringComposer::compose(std::ostream& out, const MultiInitialStateNFA& automaton) const {
	out << "MISNFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialStates().find(state) != automaton.getInitialStates().end()) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

void AutomatonToStringComposer::compose(std::ostream& out, const EpsilonNFA& automaton) const {
	out << "ENFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << " #E";
	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

void AutomatonToStringComposer::compose(std::ostream& output, const Automaton& automaton) const {
	automaton.getData().Accept((void*) &output, *this);
}

void AutomatonToStringComposer::Visit(void* data, const EpsilonNFA& automaton) const {
	this->compose(*((std::ostream*) data), automaton);
}

void AutomatonToStringComposer::Visit(void* data, const MultiInitialStateNFA& automaton) const {
	this->compose(*((std::ostream*) data), automaton);
}

void AutomatonToStringComposer::Visit(void* data, const NFA& automaton) const {
	this->compose(*((std::ostream*) data), automaton);
}

void AutomatonToStringComposer::Visit(void* data, const DFA& automaton) const {
	this->compose(*((std::ostream*) data), automaton);
}

void AutomatonToStringComposer::Visit(void*, const ExtendedNFA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const CompactNFA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const DPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const InputDrivenDPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const NPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException();
}

void AutomatonToStringComposer::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException();
}

} /* namespace automaton */
