/*
 * CompactNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef COMPACT_DFA_H_
#define COMPACT_DFA_H_

#include <map>
#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/InputAlphabet.h"
#include "../../string/LinearString.h"
#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class CompactNFA : public std::acceptor<CompactNFA, VisitableAutomatonBase, std::acceptor<CompactNFA, alib::VisitableObjectBase, AutomatonBase> >, public SingleInitialState, public InputAlphabet {
protected:
	std::map<std::pair<State, string::LinearString>, std::set<State> > transitions;
public:
	explicit CompactNFA(const State& initialState);
	explicit CompactNFA(const EpsilonNFA& other);
	explicit CompactNFA(const MultiInitialStateNFA& other);
	explicit CompactNFA(const NFA& other);
	explicit CompactNFA(const DFA& other);
	
	virtual AutomatonBase* clone() const;
	
	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);
	
	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition(const State& current, const string::LinearString& input, const State& next);

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition(const State& current, const string::LinearString& input, const State& next);

	/**
	 * @return automaton transitions
	 */
	const std::map<std::pair<State, string::LinearString>, std::set<State>>& getTransitions() const;

	/**
	 * @return automaton transitions from state
	 */
	std::map<std::pair<State, string::LinearString>, std::set<State>> getTransitionsFromState(const State& from) const;
	
	/**
	 * @return automaton transitions to state
	 */
	std::map<std::pair<State, string::LinearString>, std::set<State>> getTransitionsToState(const State& from) const;

	virtual int compare(const ObjectBase& other) const {
		return -other.compare(*this);
	}

	virtual int compare(const CompactNFA& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual explicit operator std::string() const;

	virtual int selfTypeId() const {
		return typeId<CompactNFA>();
	}
};

} /* namespace automaton */

namespace std {

template<>
struct compare<automaton::CompactNFA> {
	int operator()(const automaton::CompactNFA& first, const automaton::CompactNFA& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* COMPACT_DFA_H_ */
