/*
 * DPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "DPDA.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <algorithm>
#include <sstream>

namespace automaton {

DPDA::DPDA(const State& initialState, const alphabet::Symbol& initialPushdownSymbol) : SingleInitialSymbolPushdownStoreAlphabet(initialPushdownSymbol), SingleInitialState(initialState) {

}

AutomatonBase* DPDA::clone() const {
	return new DPDA(*this);
}

AutomatonBase* DPDA::plunder() && {
	return new DPDA(std::move(*this));
}

bool DPDA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		if (state == std::get<0>(transition.first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		if( transition.second.first == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool DPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		if (std::get<1>(transition.first).is<alphabet::Symbol>() && symbol == std::get<1>(transition.first).get<alphabet::Symbol>())
			throw AutomatonException("Symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return inputAlphabet.erase(symbol);
}

bool DPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		for (const alphabet::Symbol& popSymbol : std::get<2>(transition.first)) {
			if (symbol == popSymbol)
				throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
		}
		if (std::find(transition.second.second.begin(), transition.second.second.end(), symbol) != transition.second.second.end())
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(initialSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

bool DPDA::addTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const std::vector<alphabet::Symbol>& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	for(const alphabet::Symbol& popSymbol : pop) {
		if (stackAlphabet.find(popSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) popSymbol + "\" doesn't exist.");
		}
	}

	for(const alphabet::Symbol& pushSymbol : push) {
		if (stackAlphabet.find(pushSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) pushSymbol + "\" doesn't exist.");
		}
	}

	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> > key(from, input, pop);
	std::pair<automaton::State, std::vector<alphabet::Symbol> > value = std::make_pair(to, push);

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else if (input.is<alphabet::Symbol>())
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input.get<alphabet::Symbol>() + "\", \"pop\") -> ?? already exists.");
		else
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input.get<string::Epsilon>() + "\", \"pop\") -> ?? already exists.");
	}

	if(input.is<string::Epsilon>()) {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<automaton::State, std::vector<alphabet::Symbol> > > & transition) {
				if(std::get<0>(transition.first) == from) {
					const std::vector<alphabet::Symbol>& alpha = std::get<2>(transition.first);
					const std::vector<alphabet::Symbol>& beta = pop;

					const std::vector<alphabet::Symbol>& shorter = (alpha.size() < beta.size()) ? alpha : beta;
					const std::vector<alphabet::Symbol>& longer = (alpha.size() < beta.size()) ? beta : alpha;

					for(auto itS = shorter.begin(), itL = longer.begin(); itS != shorter.end(); itS++, itL++) {
						if(*itS != *itL) return false;
					}
					return true;
				}
				return false;
			}))
				throw exception::AlibException("Conflicting transition");
	} else {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<automaton::State, std::vector<alphabet::Symbol> > > & transition) {
				if(std::get<0>(transition.first) == from && ( std::get<1>(transition.first) == input || std::get<1>(transition.first).is<string::Epsilon>() )) {
					const std::vector<alphabet::Symbol>& alpha = std::get<2>(transition.first);
					const std::vector<alphabet::Symbol>& beta = pop;

					const std::vector<alphabet::Symbol>& shorter = (alpha.size() < beta.size()) ? alpha : beta;
					const std::vector<alphabet::Symbol>& longer = (alpha.size() < beta.size()) ? beta : alpha;

					for(auto itS = shorter.begin(), itL = longer.begin(); itS != shorter.end(); itS++, itL++) {
						if(*itS != *itL) return false;
					}
					return true;
				}
				return false;
			}))
				throw exception::AlibException("Conflicting transition");
	}

	transitions.insert(std::make_pair(key, value) );
	return true;
}

bool DPDA::addTransition(const State& from, const alphabet::Symbol& input, const std::vector<alphabet::Symbol>& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return addTransition(from, inputVariant, pop, to, push);
}

bool DPDA::addTransition(const State& from, const std::vector<alphabet::Symbol>& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addTransition(from, inputVariant, pop, to, push);
}

bool DPDA::removeTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const std::vector<alphabet::Symbol>& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> > key(from, input, pop);
	std::pair<automaton::State, std::vector<alphabet::Symbol> > value = std::make_pair(to, push);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != value) {
		if (input.is<alphabet::Symbol>())
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input.get<alphabet::Symbol>() + "\") -> \"to\" doesn't exist.");
		else
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input.get<string::Epsilon>() + "\") -> \"to\" doesn't exist.");
	}

	transitions.erase(key);
	return true;
}

bool DPDA::removeTransition(const State& from, const alphabet::Symbol& input, const std::vector<alphabet::Symbol>& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

bool DPDA::removeTransition(const State& from, const std::vector<alphabet::Symbol>& pop, const State& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return removeTransition(from, inputVariant, pop, to, push);
}

const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > >& DPDA::getTransitions() const {
	return transitions;
}

std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > > DPDA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > > transitionsFromState;
	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		if (std::get<0>(transition.first) == from) {
			transitionsFromState.insert(std::make_pair(transition.first, transition.second));
		}
	}

	return transitionsFromState;
}

std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > > DPDA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > > transitionsToState;
	for (const std::pair<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<State, std::vector<alphabet::Symbol> > >& transition : transitions) {
		if (transition.second.first == to) {
			transitionsToState.insert(std::make_pair(transition.first, transition.second));
		}
	}

	return transitionsToState;
}

int DPDA::compare(const DPDA& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, stackAlphabet, initialSymbol, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.stackAlphabet, other.initialSymbol, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void DPDA::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << states
		<< "inputAlphabet = " << inputAlphabet
		<< "initialState = " << initialState
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "initialSymbol = " << initialSymbol
		<< "transitions = " << transitions
		<< ")";
}

DPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace automaton */
