/*
 * INPUT_DRIVEN_DPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef INPUT_DRIVEN_DPDA_H_
#define INPUT_DRIVEN_DPDA_H_

#include "../../std/map.hpp"
#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/InputAlphabet.h"
#include "../common/SingleInitialSymbolPushdownStoreAlphabet.h"
#include "../../alphabet/Symbol.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputDrivenDPDA : public std::acceptor<InputDrivenDPDA, VisitableAutomatonBase, std::acceptor<InputDrivenDPDA, alib::VisitableObjectBase, AutomatonBase> >, public SingleInitialSymbolPushdownStoreAlphabet, public SingleInitialState, public InputAlphabet {
protected:
	std::map<std::pair<State, alphabet::Symbol>, State> transitions;
	std::map<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>> inputSymbolToPushdownStoreOperation;
public:
	explicit InputDrivenDPDA(const State& initialState, const alphabet::Symbol& initialPushdownSymbol);

	virtual AutomatonBase* clone() const;

	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);

	/*
	 * @copydoc Automaton::removeStackSymbol(const Symbol&)
	 */
	virtual bool removeStackSymbol(const alphabet::Symbol& symbol);

	bool setPushdownStoreOperation(const alphabet::Symbol& input, const std::vector<alphabet::Symbol>& pop, const std::vector<alphabet::Symbol>& push);

	void setPushdownStoreOperations(const std::map<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>>& operations);

	bool clearPushdownStoreOperation(const alphabet::Symbol& input);

	const std::map<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>>& getPushdownStoreOperations() const;

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition(const State& current, const alphabet::Symbol& input, const State& next);

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition(const State& current, const alphabet::Symbol& input, const State& next);

	/**
	 * @return automaton transitions
	 */
	const std::map<std::pair<State, alphabet::Symbol>, State>& getTransitions() const;

	std::map<std::pair<State, alphabet::Symbol>, State> getTransitionsFromState(const State& from) const;

	std::map<std::pair<State, alphabet::Symbol>, State> getTransitionsToState(const State& to) const;

	/**
	 * Determines whether InputDrivenDPDA is deterministic.
	 * FA is deterministic if and only if:
	 * \li \c contains only 1 initial state.
	 * \li \c is epsilon free. Trivial for this class
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic() const;

	virtual int compare(const ObjectBase& other) const {
		return -other.compare(*this);
	}

	virtual int compare(const InputDrivenDPDA& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual explicit operator std::string() const;

	virtual int selfTypeId() const {
		return typeId<InputDrivenDPDA>();
	}
};

} /* namespace automaton */

namespace std {

template<>
struct compare<automaton::InputDrivenDPDA> {
	int operator()(const automaton::InputDrivenDPDA& first, const automaton::InputDrivenDPDA& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* INPUT_DRIVEN_DPDA_H_ */
