/*
 * OneTapeDTM.cpp
 *
 *  Created on: Apr 24, 2013
 *      Author: Martin Zak
 */

#include "OneTapeDTM.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <sstream>

namespace automaton {

OneTapeDTM::OneTapeDTM(const State& initialState, const alphabet::Symbol& blank) : SingleInitialState(initialState), BlankSymbolInputTapeAlphabet(blank) {
}

AutomatonBase* OneTapeDTM::clone() const {
	return new OneTapeDTM(*this);
}

AutomatonBase* OneTapeDTM::plunder() && {
	return new OneTapeDTM(std::move(*this));
}

bool OneTapeDTM::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (std::map<std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >::const_iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
		if (state == transition->first.first || state == std::get<0>(transition->second))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in a transition.");
	}

	return states.erase(state);
}

bool OneTapeDTM::removeInputSymbol(const alphabet::Symbol& symbol) {
	return inputAlphabet.erase(symbol);
}

bool OneTapeDTM::removeTapeSymbol(const alphabet::Symbol& symbol) {
	if (inputAlphabet.find(symbol) != inputAlphabet.end()) {
		throw AutomatonException("Tape symbol \"" + (std::string) symbol + "\" is in input alphabet.");
	}

	for (std::map<std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >::const_iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
		if (symbol == transition->first.second || symbol == std::get<1>(transition->second))
			throw AutomatonException("Tape symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return inputAlphabet.erase(symbol);
}

bool OneTapeDTM::addTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& output, const Shift& shift) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (input != blankSymbol)
		if (tapeAlphabet.find(input) == tapeAlphabet.end()) {
			throw AutomatonException("Tape symbol \"" + (std::string) input + "\" doesn't exist.");
		}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (input != blankSymbol)
		if (tapeAlphabet.find(output) == tapeAlphabet.end()) {
			throw AutomatonException("Tape symbol  \"" + (std::string) output + "\" doesn't exist.");
		}

	if(shift == Shift::NOT_SET)
		throw AutomatonException("Invalid shift.");

	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);
	
	std::tuple<State, alphabet::Symbol, Shift > value(to, output, shift);

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input + "\") -> ? already exists.");
	}

	transitions.insert(std::make_pair(key, value));
	return true;
}

bool OneTapeDTM::removeTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& output, const Shift& shift) {
	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);
  
	if (transitions.find(key) == transitions.end())
		return false;
	
	std::tuple<State, alphabet::Symbol, Shift > value(to, output, shift);
	if(transitions.find(key)->second != value) 
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input + "\") -> ? doesn't exists.");

	transitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >& OneTapeDTM::getTransitions() const {
	return transitions;
}

int OneTapeDTM::compare(const OneTapeDTM& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, tapeAlphabet, blankSymbol, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.tapeAlphabet, other.blankSymbol, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void OneTapeDTM::operator>>(std::ostream& out) const {
	out << "(OneTapeDTM "
			<< "states = " << states
			<< "inputAlphabet = " << inputAlphabet
			<< "initialState = " << initialState
			<< "finalStates = " << finalStates
			<< "tapeAlphabet = " << tapeAlphabet
			<< "blankSymbol = " << blankSymbol
			<< "transitions = " << transitions
			<< ")";
}

OneTapeDTM::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace automaton */
