/*
 * OneTapeDTM.h
 *
 *  Created on: Apr 24, 2013
 *      Author: Martin Zak
 */

#ifndef ONE_TAPE_DTM_H_
#define ONE_TAPE_DTM_H_

#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/BlankSymbolInputTapeAlphabet.h"
#include "../common/Shift.h"

#include <map>
#include <set>
#include <tuple>

namespace automaton {

/**
 * One tape turing machine
 */
class OneTapeDTM : public std::acceptor<OneTapeDTM, VisitableAutomatonBase, std::acceptor<OneTapeDTM, alib::VisitableObjectBase, AutomatonBase> >, public SingleInitialState, public BlankSymbolInputTapeAlphabet {
protected:
	std::map<std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> > transitions;

public:
	explicit OneTapeDTM(const State& initialState, const alphabet::Symbol& blank);

	virtual AutomatonBase* clone() const;

	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * Removes input symbol from the input alphabet.
	 * @param symbol Symbol to remove
	 * @throws AutomatonException when symbol is part of the transition
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);

	/**
	 * Removes symbol from the tape alphabet.
	 * @param symbol Symbol to remove
	 * @throw AutomatonException when Symbol is not present in the tape alphabet,
	 * when it is used in transition or when it is present in input alphabet
	 */
	virtual bool removeTapeSymbol(const alphabet::Symbol& symbol);

	/**
	 * Adds transition to the automaton.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the TM (state, tape symbol) or when transition already exists
	 */
	bool addTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& output, const Shift& shift);

	/**
	 * Removes the transition from the TM.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the TM
	 */
	bool removeTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& output, const Shift& shift);

	/**
	 * @return TM transitions
	 */
	const std::map<std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >& getTransitions() const;

	virtual int compare(const ObjectBase& other) const {
		return -other.compare(*this);
	}

	virtual int compare(const OneTapeDTM& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual explicit operator std::string() const;

	virtual int selfTypeId() const {
		return typeId<OneTapeDTM>();
	}
};

} /* namespace automaton */

namespace std {

template<>
struct compare<automaton::OneTapeDTM> {
	int operator()(const automaton::OneTapeDTM& first, const automaton::OneTapeDTM& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* ONE_TAPE_DTM_H_ */
