/*
 * GrammarFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "GrammarFromStringParser.h"

#include "../exception/AlibException.h"

#include "Regular/RightRG.h"
#include "Regular/LeftRG.h"
#include "Regular/RightLG.h"
#include "Regular/LeftLG.h"

#include "ContextFree/LG.h"
#include "ContextFree/CFG.h"
#include "ContextFree/EpsilonFreeCFG.h"
#include "ContextFree/GNF.h"
#include "ContextFree/CNF.h"

#include "ContextSensitive/CSG.h"
#include "ContextSensitive/NonContractingGrammar.h"

#include "Unrestricted/ContextPreservingUnrestrictedGrammar.h"
#include "Unrestricted/UnrestrictedGrammar.h"

#include "../StringApi.hpp"

namespace grammar {

Grammar GrammarFromStringParser::parseGrammar(std::istream& input) const {
	return parseGrammar(input, std::set<FEATURES>({FEATURES::LEFT_LG, FEATURES::LEFT_RG, FEATURES::RIGHT_LG, FEATURES::RIGHT_RG, FEATURES::LG, FEATURES::CFG, FEATURES::EPSILON_FREE_CFG, FEATURES::CNF, FEATURES::GNF, FEATURES::CSG, FEATURES::NON_CONTRACTING_GRAMMAR, FEATURES::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR, FEATURES::UNRESTRICTED_GRAMMAR}));
}

Grammar GrammarFromStringParser::parseGrammar(std::istream& input, const std::set<FEATURES>& features) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type == GrammarFromStringLexer::TokenType::LEFT_RG) {
		if(!features.count(FEATURES::LEFT_RG)) throw exception::AlibException("Disabled Formalism LeftRG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseLeftRG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::RIGHT_RG) {
		if(!features.count(FEATURES::RIGHT_RG)) throw exception::AlibException("Disabled Formalism RightRG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseRightRG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::LEFT_LG) {
		if(!features.count(FEATURES::LEFT_LG)) throw exception::AlibException("Disabled Formalism LeftLG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseLeftLG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::RIGHT_LG) {
		if(!features.count(FEATURES::RIGHT_LG)) throw exception::AlibException("Disabled Formalism RightLG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseRightLG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::LG) {
		if(!features.count(FEATURES::LG)) throw exception::AlibException("Disabled formalism LG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseLG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CFG) {
		if(!features.count(FEATURES::CFG)) throw exception::AlibException("Disabled formalism CFG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseCFG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::EPSILON_FREE_CFG) {
		if(!features.count(FEATURES::EPSILON_FREE_CFG)) throw exception::AlibException("Disabled formalism EpsilonFreeCFG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseEpsilonFreeCFG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::GNF) {
		if(!features.count(FEATURES::GNF)) throw exception::AlibException("Disabled formalism GNF");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseGNF(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CNF) {
		if(!features.count(FEATURES::CNF)) throw exception::AlibException("Disabled formalism CNF");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseCNF(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::NON_CONTRACTING_GRAMMAR) {
		if(!features.count(FEATURES::NON_CONTRACTING_GRAMMAR)) throw exception::AlibException("Disabled formalism NonContractingGrammar");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseNonContractingGrammar(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CSG) {
		if(!features.count(FEATURES::CSG)) throw exception::AlibException("Disabled formalism CSG");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseCSG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR) {
		if(!features.count(FEATURES::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR)) throw exception::AlibException("Disabled formalism ContextPreservingUnrestrictedGrammar");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseContextPreservingUnrestrictedGrammar(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::UNRESTRICTED_GRAMMAR) {
		if(!features.count(FEATURES::UNRESTRICTED_GRAMMAR)) throw exception::AlibException("Disabled formalism UnrestrictedGrammar");
		m_GrammarLexer.putback(input, token);
		return Grammar(parseUnrestrictedGrammar(input));
	} else {
		throw exception::AlibException("Formalism not recognised (token = \"" + token.value + "\")");
	}
}

std::set<alphabet::Symbol> GrammarFromStringParser::parseSet(std::istream& input) const {
	std::set<alphabet::Symbol> res;

	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::AlibException();
	}

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		m_GrammarLexer.putback(input, token);
		while(true) {
			alphabet::Symbol symbol = alib::stringApi<alphabet::Symbol>::parse(input);
			res.insert(symbol);

			token = m_GrammarLexer.next(input);
			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::AlibException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::AlibException("Expected SET_END token");
	}

	return res;
}

std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> GrammarFromStringParser::parseCFLikeRules(std::istream& input) const {
	std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> result;

	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::AlibException();
	}

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		m_GrammarLexer.putback(input, token);
		while(true) {
			alphabet::Symbol lhs = alib::stringApi<alphabet::Symbol>::parse(input);

			token = m_GrammarLexer.next(input);
			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) {
				throw exception::AlibException();
			}

			while(true) {
				std::vector<alphabet::Symbol> rhs;

				token = m_GrammarLexer.next(input);
				if(token.type != GrammarFromStringLexer::TokenType::COMMA && token.type != GrammarFromStringLexer::TokenType::SET_END && token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
					m_GrammarLexer.putback(input, token);

					rhs.push_back(alib::stringApi<alphabet::Symbol>::parse(input));
					token = m_GrammarLexer.next(input);
					if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR || token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
						break;
					}
				}
				result[lhs].insert(rhs);
				if(token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
					break;
				}
				if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
					throw exception::AlibException("Expected SEPARATOR, SETEND or COMMA token");
				}
			}

			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::AlibException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::AlibException("Expected SET_END token");
	}
	return result;
}

template<class T>
T GrammarFromStringParser::parseCFLikeGrammar(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		throw exception::AlibException("Unrecognised Tuple begin token.");
	}

	std::set<alphabet::Symbol> nonterminals = parseSet(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	std::set<alphabet::Symbol> terminals = parseSet(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> rules = parseCFLikeRules(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	alphabet::Symbol initialSymbol = alib::stringApi<alphabet::Symbol>::parse(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_END) {
		throw exception::AlibException("Unrecognised Tuple end token.");
	}

	T grammar(nonterminals, terminals, initialSymbol);
	for(const auto& rule : rules) {
		for(const auto& ruleRHS : rule.second) {
			grammar.addRawRule(rule.first, ruleRHS);
		}
	}
	return grammar;
}

std::map<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>> GrammarFromStringParser::parseCSLikeRules(std::istream& input) const {
	std::map<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>> result;

	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::AlibException();
	}

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		m_GrammarLexer.putback(input, token);
		while(true) {
			std::vector<alphabet::Symbol> lhs;
			while(true) {
				lhs.push_back(alib::stringApi<alphabet::Symbol>::parse(input));
				token = m_GrammarLexer.next(input);
				if(token.type == GrammarFromStringLexer::TokenType::MAPS_TO) {
					break;
				}
				m_GrammarLexer.putback(input, token);
			}

			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) {
				throw exception::AlibException();
			}

			while(true) {
				std::vector<alphabet::Symbol> rhs;

				token = m_GrammarLexer.next(input);
				if(token.type != GrammarFromStringLexer::TokenType::COMMA && token.type != GrammarFromStringLexer::TokenType::SET_END && token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
					m_GrammarLexer.putback(input, token);

					rhs.push_back(alib::stringApi<alphabet::Symbol>::parse(input));
					token = m_GrammarLexer.next(input);
					if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR || token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
						break;
					}
				}
				result[lhs].insert(rhs);
				if(token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
					break;
				}
				if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
					throw exception::AlibException("Expected SEPARATOR, SETEND or COMMA token");
				}
			}

			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::AlibException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::AlibException("Expected SET_END token");
	}
	return result;
}

template<class T>
T GrammarFromStringParser::parseCSLikeGrammar(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		throw exception::AlibException("Unrecognised Tuple begin token.");
	}

	std::set<alphabet::Symbol> nonterminals = parseSet(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	std::set<alphabet::Symbol> terminals = parseSet(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	std::map<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>> rules = parseCSLikeRules(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	alphabet::Symbol initialSymbol = alib::stringApi<alphabet::Symbol>::parse(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_END) {
		throw exception::AlibException("Unrecognised Tuple end token.");
	}

	T grammar(nonterminals, terminals, initialSymbol);
	for(const auto& rule : rules) {
		for(const auto& ruleRHS : rule.second) {
			grammar.addRule(rule.first, ruleRHS);
		}
	}
	return grammar;
}

std::map<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>> GrammarFromStringParser::parsePreservingCSLikeRules(std::istream& input) const {
	std::map<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>> result;

	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::AlibException();
	}

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		m_GrammarLexer.putback(input, token);
		while(true) {
			std::vector<alphabet::Symbol> lContext;
			token = m_GrammarLexer.next(input);
			if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
				m_GrammarLexer.putback(input, token);
				lContext.push_back(alib::stringApi<alphabet::Symbol>::parse(input));
				token = m_GrammarLexer.next(input);
				if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR) {
					break;
				}
			}

			if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
				throw exception::AlibException();
			}

			alphabet::Symbol lhs = alib::stringApi<alphabet::Symbol>::parse(input);

			token = m_GrammarLexer.next(input);
			if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
				throw exception::AlibException();
			}

			std::vector<alphabet::Symbol> rContext;
			token = m_GrammarLexer.next(input);
			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) while(true) {
				m_GrammarLexer.putback(input, token);
				rContext.push_back(alib::stringApi<alphabet::Symbol>::parse(input));
				token = m_GrammarLexer.next(input);
				if(token.type == GrammarFromStringLexer::TokenType::MAPS_TO) {
					break;
				}
			}

			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) {
				throw exception::AlibException();
			}

			std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>> key = std::make_tuple(lContext, lhs, rContext);

			while(true) {
				std::vector<alphabet::Symbol> rhs;

				token = m_GrammarLexer.next(input);
				if(token.type != GrammarFromStringLexer::TokenType::COMMA && token.type != GrammarFromStringLexer::TokenType::SET_END && token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
					m_GrammarLexer.putback(input, token);

					rhs.push_back(alib::stringApi<alphabet::Symbol>::parse(input));
					token = m_GrammarLexer.next(input);
					if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR || token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
						break;
					}
				}
				result[key].insert(rhs);
				if(token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
					break;
				}
				if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
					throw exception::AlibException("Expected SEPARATOR, SETEND or COMMA token");
				}
			}

			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::AlibException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::AlibException("Expected SET_END token");
	}
	return result;
}

template<class T>
T GrammarFromStringParser::parsePreservingCSLikeGrammar(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		throw exception::AlibException("Unrecognised Tuple begin token.");
	}

	std::set<alphabet::Symbol> nonterminals = parseSet(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	std::set<alphabet::Symbol> terminals = parseSet(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	std::map<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>> rules = parsePreservingCSLikeRules(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::AlibException("Unrecognised Comma token.");
	}

	alphabet::Symbol initialSymbol = alib::stringApi<alphabet::Symbol>::parse(input);

	token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_END) {
		throw exception::AlibException("Unrecognised Tuple end token.");
	}

	T grammar(nonterminals, terminals, initialSymbol);
	for(const auto& rule : rules) {
		for(const auto& ruleRHS : rule.second) {
			grammar.addRule(std::get<0>(rule.first), std::get<1>(rule.first), std::get<2>(rule.first), ruleRHS);
		}
	}
	return grammar;
}

RightRG GrammarFromStringParser::parseRightRG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::RIGHT_RG) {
		throw exception::AlibException("Unrecognised RightRG token.");
	}

	return parseCFLikeGrammar<RightRG>(input);
}

LeftRG GrammarFromStringParser::parseLeftRG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::LEFT_RG) {
		throw exception::AlibException("Unrecognised LeftRG token.");
	}

	return parseCFLikeGrammar<LeftRG>(input);
}

RightLG GrammarFromStringParser::parseRightLG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::RIGHT_LG) {
		throw exception::AlibException("Unrecognised RightLG token.");
	}

	return parseCFLikeGrammar<RightLG>(input);
}

LeftLG GrammarFromStringParser::parseLeftLG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::LEFT_LG) {
		throw exception::AlibException("Unrecognised LeftLG token.");
	}

	return parseCFLikeGrammar<LeftLG>(input);
}

LG GrammarFromStringParser::parseLG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::LG) {
		throw exception::AlibException("Unrecognised LG token.");
	}

	return parseCFLikeGrammar<LG>(input);
}

CFG GrammarFromStringParser::parseCFG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CFG) {
		throw exception::AlibException("Unrecognised CFG token.");
	}

	return parseCFLikeGrammar<CFG>(input);
}

EpsilonFreeCFG GrammarFromStringParser::parseEpsilonFreeCFG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::EPSILON_FREE_CFG) {
		throw exception::AlibException("Unrecognised EpsilonFreeCFG token.");
	}

	return parseCFLikeGrammar<EpsilonFreeCFG>(input);
}

GNF GrammarFromStringParser::parseGNF(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::GNF) {
		throw exception::AlibException("Unrecognised GNF token.");
	}

	return parseCFLikeGrammar<GNF>(input);
}

CNF GrammarFromStringParser::parseCNF(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CNF) {
		throw exception::AlibException("Unrecognised CNF token.");
	}

	return parseCFLikeGrammar<CNF>(input);
}

NonContractingGrammar GrammarFromStringParser::parseNonContractingGrammar(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::NON_CONTRACTING_GRAMMAR) {
		throw exception::AlibException("Unrecognised NonContractingGrammar token.");
	}

	return parseCSLikeGrammar<NonContractingGrammar>(input);
}

CSG GrammarFromStringParser::parseCSG(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CSG) {
		throw exception::AlibException("Unrecognised CSG token.");
	}

	return parsePreservingCSLikeGrammar<CSG>(input);
}

ContextPreservingUnrestrictedGrammar GrammarFromStringParser::parseContextPreservingUnrestrictedGrammar(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR) {
		throw exception::AlibException("Unrecognised ContextPreservingUnrestrictedGrammar token.");
	}

	return parsePreservingCSLikeGrammar<ContextPreservingUnrestrictedGrammar>(input);
}

UnrestrictedGrammar GrammarFromStringParser::parseUnrestrictedGrammar(std::istream& input) const {
	GrammarFromStringLexer::Token token = m_GrammarLexer.next(input);
	if(token.type != GrammarFromStringLexer::TokenType::UNRESTRICTED_GRAMMAR) {
		throw exception::AlibException("Unrecognised UnrestrictedGrammar token.");
	}

	return parseCSLikeGrammar<UnrestrictedGrammar>(input);
}

} /* namespace grammar */
