/*
 * RightRG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "RightRG.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

namespace grammar {

RightRG::RightRG(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {

}

RightRG::RightRG(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* RightRG::clone() const {
	return new RightRG(*this);
}

GrammarBase* RightRG::plunder() && {
	return new RightRG(std::move(*this));
}

bool RightRG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>>& rule : rules) {
		for(const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rhs : rule.second) {
			if((rhs.is<alphabet::Symbol>() && rhs.get<alphabet::Symbol>() == symbol) || (rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>().first == symbol))
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
		}
	}

	return terminalAlphabet.erase(symbol);
}

bool RightRG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>>& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rhs : rule.second)
			if(rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>().second == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");

	return nonterminalAlphabet.erase(symbol);
}

bool RightRG::addRule(const alphabet::Symbol& leftHandSide, const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	if(rightHandSide.is<alphabet::Symbol>()) {
		const alphabet::Symbol& rhs = rightHandSide.get<alphabet::Symbol>();

		if(!terminalAlphabet.count(rhs))
			throw GrammarException("Rule must rewrite to terminal symbol");

		return rules[leftHandSide].insert(rightHandSide).second;
	} else {
		const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = rightHandSide.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();

		if(!terminalAlphabet.count(rhs.first) || !nonterminalAlphabet.count(rhs.second))
			throw GrammarException("Rule must rewrite to terminal symbol followed by nonterminal symbol");

		return rules[leftHandSide].insert(rightHandSide).second;
	}
}

bool RightRG::addRule(const alphabet::Symbol& leftHandSide, const alphabet::Symbol& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

bool RightRG::addRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, alphabet::Symbol>& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

const std::map<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>> & RightRG::getRules() const {
	return rules;
}

bool RightRG::removeRule(const alphabet::Symbol& leftHandSide, const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

bool RightRG::removeRule(const alphabet::Symbol& leftHandSide, const alphabet::Symbol& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool RightRG::removeRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, alphabet::Symbol>& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool RightRG::addRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		if(leftHandSide != initialSymbol) throw exception::AlibException("Illegal left hand side of epsilon rule");
		bool res = getGeneratesEpsilon();
		setGeneratesEpsilon(true);
		return res;
	} else if(rightHandSide.size() == 1) {
		return addRule(leftHandSide, rightHandSide[0]);
	} else if(rightHandSide.size() == 2) {
		return addRule(leftHandSide, std::make_pair(rightHandSide[0], rightHandSide[1]));
	} else {
		throw exception::AlibException("Invalid right hand side");
	}
}

std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> RightRG::getRawRules() const {
	std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> res;
	for(const auto& rule : getRules()) {
		for(const auto& rhs : rule.second) {
			if(rhs.is<alphabet::Symbol>()) {
				std::vector<alphabet::Symbol> tmp;
				tmp.push_back(rhs.get<alphabet::Symbol>());
				res[rule.first].insert(tmp);
			} else {
				const auto& rhsPair = rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
				std::vector<alphabet::Symbol> tmp;
				tmp.push_back(rhsPair.first);
				tmp.push_back(rhsPair.second);
				res[rule.first].insert(tmp);
			}
		}
	}
	if(getGeneratesEpsilon()) {
		res[getInitialSymbol()].insert(std::vector<alphabet::Symbol> {});
	}
	return res;
}

bool RightRG::removeRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		if(leftHandSide != initialSymbol) throw exception::AlibException("Illegal left hand side of epsilon rule");
		bool res = getGeneratesEpsilon();
		setGeneratesEpsilon(false);
		return res;
	} else if(rightHandSide.size() == 1) {
		return removeRule(leftHandSide, rightHandSide[0]);
	} else if(rightHandSide.size() == 2) {
		return removeRule(leftHandSide, std::make_pair(rightHandSide[0], rightHandSide[1]));
	} else {
		throw exception::AlibException("Invalid right hand side");
	}
}

void RightRG::setGeneratesEpsilon(bool genEps) {
	generatesEpsilon = genEps;
}

bool RightRG::getGeneratesEpsilon() const {
	return generatesEpsilon;
}

int RightRG::compare(const RightRG& other) const {
	auto first = std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules);
	auto second = std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void RightRG::operator>>(std::ostream& out) const {
	out << "(RightRG "
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< "generatesEpsilon = " << generatesEpsilon
			<< ")";
}

RightRG::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace grammar */
