#include "GraphFromXMLParser.h"
#include "../sax/ParserException.h"

#include "../XmlApi.hpp"

namespace graph {

bool GraphFromXMLParser::first(const std::list<sax::Token> &input) const
{
	if (alib::xmlApi<DirectedGraph>::first(input)) {
		return true;
	}

	return false;
}

Graph GraphFromXMLParser::parseGraph(std::list<sax::Token> &input) const
{
	if (alib::xmlApi<DirectedGraph>::first(input)) {
		return Graph(parseDirectedGraph(input));
	} else if (alib::xmlApi<UndirectedGraph>::first(input)) {
		return Graph(parseUndirectedGraph(input));
	}

	throw exception::AlibException();
}

DirectedGraph GraphFromXMLParser::parseDirectedGraph(std::list<sax::Token> &input) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::GRAPH_DIRECTED_GRAPH);

	DirectedGraph graph(parseRepresentation(input));
	parseNodes(input, graph);
	parseDirectedEdges(input, graph);
	parseNodeValues(input, graph);
	parseDirectedEdgeValues(input, graph);

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::GRAPH_DIRECTED_GRAPH);
	return graph;
}

UndirectedGraph GraphFromXMLParser::parseUndirectedGraph(std::list<sax::Token> &input) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::GRAPH_UNDIRECTED_GRAPH);

	UndirectedGraph graph(parseRepresentation(input));
	parseNodes(input, graph);
	parseUndirectedEdges(input, graph);
	parseNodeValues(input, graph);
	parseUndirectedEdgeValues(input, graph);

	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::GRAPH_UNDIRECTED_GRAPH);
	return graph;
}

REPRESENTATION GraphFromXMLParser::parseRepresentation(std::list<sax::Token> &input) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "representation");

	REPRESENTATION r = representationFromString(popTokenData(input, sax::Token::TokenType::CHARACTER));

	popToken(input, sax::Token::TokenType::END_ELEMENT, "representation");
	return r;
}

Node GraphFromXMLParser::parseNode(std::list<sax::Token> &input) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "node");

	Node n = Node(alib::xmlApi<label::Label>::parse(input));

	popToken(input, sax::Token::TokenType::END_ELEMENT, "node");
	return n;
}

DirectedEdge GraphFromXMLParser::parseDirectedEdge(std::list<sax::Token> &input) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "edge");

	popToken(input, sax::Token::TokenType::START_ELEMENT, "from");
	Node from = parseNode(input);
	popToken(input, sax::Token::TokenType::END_ELEMENT, "from");

	label::Label name = alib::xmlApi<label::Label>::parse(input);

	popToken(input, sax::Token::TokenType::START_ELEMENT, "to");
	Node to = parseNode(input);
	popToken(input, sax::Token::TokenType::END_ELEMENT, "to");

	popToken(input, sax::Token::TokenType::END_ELEMENT, "edge");
	return DirectedEdge(from, to, name);
}

UndirectedEdge GraphFromXMLParser::parseUndirectedEdge(std::list<sax::Token> &input) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "edge");

	popToken(input, sax::Token::TokenType::START_ELEMENT, "first");
	Node first = parseNode(input);
	popToken(input, sax::Token::TokenType::END_ELEMENT, "first");

	label::Label name = alib::xmlApi<label::Label>::parse(input);

	popToken(input, sax::Token::TokenType::START_ELEMENT, "second");
	Node second = parseNode(input);
	popToken(input, sax::Token::TokenType::END_ELEMENT, "second");

	popToken(input, sax::Token::TokenType::END_ELEMENT, "edge");
	return UndirectedEdge(first, second, name);
}

int GraphFromXMLParser::parseValue(std::list<sax::Token> &input) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "value");

	int value = std::stoi(popTokenData(input, sax::Token::TokenType::CHARACTER));

	popToken(input, sax::Token::TokenType::END_ELEMENT, "value");
	return value;
}

template<typename T>
void GraphFromXMLParser::parseNodes(std::list<sax::Token> &input, T &graph) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "nodes");

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		graph.addNode(parseNode(input));
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "nodes");
}

void GraphFromXMLParser::parseDirectedEdges(std::list<sax::Token> &input, DirectedGraph &graph) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "edges");

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		graph.addEdge(parseDirectedEdge(input));
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "edges");
}

void GraphFromXMLParser::parseUndirectedEdges(std::list<sax::Token> &input, UndirectedGraph &graph) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "edges");

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		graph.addEdge(parseUndirectedEdge(input));
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "edges");
}

template<typename T>
void GraphFromXMLParser::parseNodeValues(std::list<sax::Token> &input, T &graph) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "nodevalues");

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		graph::Node node = parseNode(input);
		int value = parseValue(input);

		graph.setNodeValue(node, value);
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "nodevalues");
}

void GraphFromXMLParser::parseDirectedEdgeValues(std::list<sax::Token> &input, DirectedGraph &graph) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "edgevalues");

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		graph::DirectedEdge edge = parseDirectedEdge(input);
		int value = parseValue(input);

		graph.setEdgeValue(edge, value);
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "edgevalues");
}

void GraphFromXMLParser::parseUndirectedEdgeValues(std::list<sax::Token> &input, UndirectedGraph &graph) const
{
	popToken(input, sax::Token::TokenType::START_ELEMENT, "edgevalues");

	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		graph::UndirectedEdge edge = parseUndirectedEdge(input);
		int value = parseValue(input);

		graph.setEdgeValue(edge, value);
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "edgevalues");
}

} // namespace graph

