#include "GraphToStringComposer.h"
#include "../StringApi.hpp"

#include <algorithm>

namespace graph {

void GraphToStringComposer::compose(std::ostream &out, const Graph &graph) const
{
	out << "(";
	graph.getData().Accept(static_cast<void*>(&out), *this);
	out << ")";
}

void GraphToStringComposer::Visit(void *data, const DirectedGraph &graph) const
{
	std::ostream &out = *static_cast<std::ostream*>(data);

	out << "DirectedGraph:";
	composeRepresentation(out, graph.getRepresentation());
	out << ":";
	composeNodes(out, graph.getNodes());
	out << ":";
	composeDirectedEdges(out, graph.getEdges());
	out << ":";
	composeNodeValues(out, graph);
	out << ":";
	composeEdgeValues(out, graph);
}

void GraphToStringComposer::Visit(void *data, const UndirectedGraph &graph) const
{
	std::ostream &out = *static_cast<std::ostream*>(data);

	out << "UndirectedGraph:";
	composeRepresentation(out, graph.getRepresentation());
	out << ":";
	composeNodes(out, graph.getNodes());
	out << ":";
	composeUndirectedEdges(out, graph.getEdges());
	out << ":";
	composeNodeValues(out, graph);
	out << ":";
	composeEdgeValues(out, graph);
}

void GraphToStringComposer::Visit(void *data, const Node &node) const
{
	std::ostream &out = *static_cast<std::ostream*>(data);

	alib::stringApi<label::Label>::compose(out, node.getName());
}

void GraphToStringComposer::Visit(void *data, const DirectedEdge &edge) const
{
	std::ostream &out = *static_cast<std::ostream*>(data);

	edge.getFromNode().Accept(static_cast<void*>(&out), *this);
	alib::stringApi<label::Label>::compose(out, edge.getName());
	edge.getToNode().Accept(static_cast<void*>(&out), *this);
}

void GraphToStringComposer::Visit(void *data, const UndirectedEdge &edge) const
{
	std::ostream &out = *static_cast<std::ostream*>(data);

	edge.getFirstNode().Accept(static_cast<void*>(&out), *this);
	alib::stringApi<label::Label>::compose(out, edge.getName());
	edge.getSecondNode().Accept(static_cast<void*>(&out), *this);
}

void GraphToStringComposer::composeRepresentation(std::ostream &out, REPRESENTATION representation) const
{
	out << representationToString(representation);
}

void GraphToStringComposer::composeNodes(std::ostream &out, const std::set<Node> &nodes) const
{
	bool first = true;

	for (const Node &node : nodes) {
		if (first) {
			first = false;
		} else {
			out << ",";
		}
		node.Accept(static_cast<void*>(&out), *this);
	}
}

void GraphToStringComposer::composeDirectedEdges(std::ostream &out, const std::set<DirectedEdge> &edges) const
{
	bool first = true;

	for (const DirectedEdge &edge : edges) {
		if (first) {
			first = false;
		} else {
			out << ",";
		}
		edge.Accept(static_cast<void*>(&out), *this);
	}
}

void GraphToStringComposer::composeUndirectedEdges(std::ostream &out, const std::set<UndirectedEdge> &edges) const
{
	bool first = true;

	for (const UndirectedEdge &edge : edges) {
		if (first) {
			first = false;
		} else {
			out << ",";
		}
		edge.Accept(static_cast<void*>(&out), *this);
	}
}

template<typename T>
void GraphToStringComposer::composeNodeValues(std::ostream &out, const T &graph) const
{
	bool first = true;

	for (auto i : graph.nodeValues) {
		if (first) {
			first = false;
		} else {
			out << ",";
		}
		i.first.Accept(static_cast<void*>(&out), *this);
		out << "=" << i.second;
	}
}

template<typename T>
void GraphToStringComposer::composeEdgeValues(std::ostream &out, const T &graph) const
{
	bool first = true;

	for (auto i : graph.edgeValues) {
		if (first) {
			first = false;
		} else {
			out << ",";
		}
		i.first.Accept(static_cast<void*>(&out), *this);
		out << "=" << i.second;
	}
}

} // namespace graph
