#include "GraphToXMLComposer.h"
#include "GraphRepresentation.h"

#include "../XmlApi.hpp"
#include "../std/itos.h"

namespace graph {

void GraphToXMLComposer::Visit(void *data, const Node &node) const
{
	std::list<sax::Token> &out = *static_cast<std::list<sax::Token>*>(data);

	out.push_back(sax::Token("node", sax::Token::TokenType::START_ELEMENT));
	alib::xmlApi<label::Label>::compose(out, node.getName());
	out.push_back(sax::Token("node", sax::Token::TokenType::END_ELEMENT));
}

void GraphToXMLComposer::Visit(void *data, const DirectedEdge &edge) const
{
	std::list<sax::Token> &out = *static_cast<std::list<sax::Token>*>(data);

	out.push_back(sax::Token("edge", sax::Token::TokenType::START_ELEMENT));

	out.push_back(sax::Token("from", sax::Token::TokenType::START_ELEMENT));
	edge.getFromNode().Accept(static_cast<void*>(&out), *this);
	out.push_back(sax::Token("from", sax::Token::TokenType::END_ELEMENT));

	alib::xmlApi<label::Label>::compose(out, edge.getName());

	out.push_back(sax::Token("to", sax::Token::TokenType::START_ELEMENT));
	edge.getToNode().Accept(static_cast<void*>(&out), *this);
	out.push_back(sax::Token("to", sax::Token::TokenType::END_ELEMENT));

	out.push_back(sax::Token("edge", sax::Token::TokenType::END_ELEMENT));
}

void GraphToXMLComposer::Visit(void *data, const UndirectedEdge &edge) const
{
	std::list<sax::Token> &out = *static_cast<std::list<sax::Token>*>(data);

	out.push_back(sax::Token("edge", sax::Token::TokenType::START_ELEMENT));

	out.push_back(sax::Token("first", sax::Token::TokenType::START_ELEMENT));
	edge.getFirstNode().Accept(static_cast<void*>(&out), *this);
	out.push_back(sax::Token("first", sax::Token::TokenType::END_ELEMENT));

	alib::xmlApi<label::Label>::compose(out, edge.getName());

	out.push_back(sax::Token("second", sax::Token::TokenType::START_ELEMENT));
	edge.getSecondNode().Accept(static_cast<void*>(&out), *this);
	out.push_back(sax::Token("second", sax::Token::TokenType::END_ELEMENT));

	out.push_back(sax::Token("edge", sax::Token::TokenType::END_ELEMENT));
}

void GraphToXMLComposer::compose(std::list<sax::Token> &out, const GraphBase &graph) const
{
	graph.Accept((void*) &out, alib::ToXMLComposers::toXMLComposers);
}

void GraphToXMLComposer::compose(std::list<sax::Token> &out, const Graph &graph) const
{
	graph.getData().Accept((void*) &out, alib::ToXMLComposers::toXMLComposers);
}

void GraphToXMLComposer::compose(std::list<sax::Token> &out, const DirectedGraph &graph) const
{
	out.push_back(sax::Token(alib::Names::GRAPH_DIRECTED_GRAPH, sax::Token::TokenType::START_ELEMENT));

	composeRepresentation(out, representationToString(graph.getRepresentation()));
	composeNodes(out, graph.getNodes());
	composeEdges(out, graph.getEdges());
	composeNodeValues(out, graph);
	composeEdgeValues(out, graph);

	out.push_back(sax::Token(alib::Names::GRAPH_DIRECTED_GRAPH, sax::Token::TokenType::END_ELEMENT));
}

void GraphToXMLComposer::compose(std::list<sax::Token> &out, const UndirectedGraph &graph) const
{
	out.push_back(sax::Token(alib::Names::GRAPH_UNDIRECTED_GRAPH, sax::Token::TokenType::START_ELEMENT));

	composeRepresentation(out, representationToString(graph.getRepresentation()));
	composeNodes(out, graph.getNodes());
	composeEdges(out, graph.getEdges());
	composeNodeValues(out, graph);
	composeEdgeValues(out, graph);

	out.push_back(sax::Token(alib::Names::GRAPH_UNDIRECTED_GRAPH, sax::Token::TokenType::END_ELEMENT));
}

void GraphToXMLComposer::composeRepresentation(std::list<sax::Token> &out, const std::string &representation) const
{
	out.push_back(sax::Token("representation", sax::Token::TokenType::START_ELEMENT));
	out.push_back(sax::Token(representation, sax::Token::TokenType::CHARACTER));
	out.push_back(sax::Token("representation", sax::Token::TokenType::END_ELEMENT));
}

void GraphToXMLComposer::composeNodes(std::list<sax::Token> &out, const std::set<Node> &nodes) const
{
	out.push_back(sax::Token("nodes", sax::Token::TokenType::START_ELEMENT));
	for (const Node &node : nodes) {
		node.Accept(static_cast<void*>(&out), *this);
	}
	out.push_back(sax::Token("nodes", sax::Token::TokenType::END_ELEMENT));
}

template<typename T>
void GraphToXMLComposer::composeEdges(std::list<sax::Token> &out, const std::set<T> &edges) const
{
	out.push_back(sax::Token("edges", sax::Token::TokenType::START_ELEMENT));
	for (const auto &edge : edges) {
		edge.Accept(static_cast<void*>(&out), *this);
	}
	out.push_back(sax::Token("edges", sax::Token::TokenType::END_ELEMENT));
}

template<typename T>
void GraphToXMLComposer::composeNodeValues(std::list<sax::Token> &out, const T &graph) const
{
	out.push_back(sax::Token("nodevalues", sax::Token::TokenType::START_ELEMENT));
	for (auto i : graph.nodeValues) {
		i.first.Accept(static_cast<void*>(&out), *this);
		out.push_back(sax::Token("value", sax::Token::TokenType::START_ELEMENT));
		out.push_back(sax::Token(std::itos(i.second), sax::Token::TokenType::CHARACTER));
		out.push_back(sax::Token("value", sax::Token::TokenType::END_ELEMENT));
	}
	out.push_back(sax::Token("nodevalues", sax::Token::TokenType::END_ELEMENT));
}

template<typename T>
void GraphToXMLComposer::composeEdgeValues(std::list<sax::Token> &out, const T &graph) const
{
	out.push_back(sax::Token("edgevalues", sax::Token::TokenType::START_ELEMENT));
	for (auto &i : graph.edgeValues) {
		i.first.Accept(static_cast<void*>(&out), *this);
		out.push_back(sax::Token("value", sax::Token::TokenType::START_ELEMENT));
		out.push_back(sax::Token(std::itos(i.second), sax::Token::TokenType::CHARACTER));
		out.push_back(sax::Token("value", sax::Token::TokenType::END_ELEMENT));
	}
	out.push_back(sax::Token("edgevalues", sax::Token::TokenType::END_ELEMENT));
}

} // namespace graph
