#include "AdjacencyListDirectedGraph.h"
#include "../../std/compare.hpp"

#include <algorithm>

namespace graph {

template<typename T>
inline void listRemoveOne(std::list<T> &lst, const T &item)
{
	lst.erase(std::find(lst.begin(), lst.end(), item));
}

std::set<Node> AdjacencyListDirectedGraph::allNodes() const
{
	std::set<Node> out;

	for (auto i : adj) {
		out.insert(i.first);
	}

	return out;
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::allEdges() const
{
	return edges;
}

bool AdjacencyListDirectedGraph::hasNode(const Node &node) const
{
	return adj.find(node) != adj.end();
}

bool AdjacencyListDirectedGraph::hasEdge(const DirectedEdge &edge) const
{
	return edges.find(edge) != edges.end();
}

bool AdjacencyListDirectedGraph::hasEdge(const Node &from, const Node &to) const
{
	auto search = adj.find(from);
	if (search == adj.end()) {
		return false;
	}
	return std::find(search->second.begin(), search->second.end(), to) != search->second.end();
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::findEdges(const Node &from, const Node &to) const
{
	std::set<DirectedEdge> out;

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == from && e.getToNode() == to) {
			out.insert(e);
		}
	}

	return out;
}

std::set<Node> AdjacencyListDirectedGraph::neighbors(const Node &node) const
{
	std::set<Node> out;

	auto search = adj.find(node);
	if (search == adj.end()) {
		return out;
	}

	std::copy(search->second.begin(), search->second.end(), std::inserter(out, out.begin()));
	return out;
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::neighborEdges(const Node &node) const
{
	std::set<DirectedEdge> out;
	const std::set<Node> &nghbrs = neighbors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == node && nghbrs.find(e.getToNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyListDirectedGraph::addNode(const Node &node)
{
	if (adj.find(node) != adj.end()) {
		return false;
	}

	adj[node]; // insert key
	return true;
}

bool AdjacencyListDirectedGraph::removeNode(const Node &node)
{
	auto search = adj.find(node);
	if (search == adj.end()) {
		return false;
	}

	adj.erase(search);

	auto i = edges.begin();
	while (i != edges.end()) {
		if (i->getFromNode() == node) {
			edges.erase(i++);
		} else if (i->getToNode() == node) {
			listRemoveOne(adj[i->getFromNode()], node);
			edges.erase(i++);
		} else {
			i++;
		}
	}

	return true;
}

bool AdjacencyListDirectedGraph::addEdge(const DirectedEdge &edge)
{
	if (edges.find(edge) != edges.end()) {
		return false;
	}

	edges.insert(edge);

	adj[edge.getFromNode()].push_back(edge.getToNode());
	adj[edge.getToNode()]; // insert key

	return true;
}

bool AdjacencyListDirectedGraph::removeEdge(const DirectedEdge &edge)
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);

	auto &lst = adj[edge.getFromNode()];
	listRemoveOne(lst, edge.getToNode());
	if (lst.empty()) {
		adj.erase(edge.getFromNode());
	}

	return true;
}

void AdjacencyListDirectedGraph::copy(IDirectedGraph *other)
{
	AdjacencyListDirectedGraph *o = static_cast<AdjacencyListDirectedGraph*>(other);

	edges = o->edges;
	adj = o->adj;
}

int AdjacencyListDirectedGraph::compare(IDirectedGraph *other) const
{
	AdjacencyListDirectedGraph *o = static_cast<AdjacencyListDirectedGraph*>(other);

	std::compare<decltype(edges)> comp;
	int res = comp(edges, o->edges);
	if (res != 0) {
		return res;
	}

	if (adj.size() < o->adj.size()) {
		return -1;
	}

	if (adj.size() > o->adj.size()) {
		return 1;
	}

	for (auto i : adj) {
		auto lst1 = i.second;
		auto lst2 = o->adj.at(i.first);

		std::set<Node> adj1;
		std::set<Node> adj2;
		std::copy(lst1.begin(), lst1.end(), std::inserter(adj1, adj1.begin()));
		std::copy(lst2.begin(), lst2.end(), std::inserter(adj2, adj2.begin()));

		std::compare<std::set<Node>> comp;
		int res = comp(adj1, adj2);
		if (res != 0) {
			return res;
		}
	}

	return 0;
}

} // namespace graph
