#include "AdjacencyMatrixDirectedGraph.h"
#include "../../std/compare.hpp"

namespace graph {

std::set<Node> AdjacencyMatrixDirectedGraph::allNodes() const
{
	std::set<Node> out;

	for (auto i : adj) {
		out.insert(i.first);
	}

	return out;
}

std::set<DirectedEdge> AdjacencyMatrixDirectedGraph::allEdges() const
{
	return edges;
}

bool AdjacencyMatrixDirectedGraph::hasNode(const Node &node) const
{
	return adj.find(node) != adj.end();
}

bool AdjacencyMatrixDirectedGraph::hasEdge(const DirectedEdge &edge) const
{
	return edges.find(edge) != edges.end();
}

bool AdjacencyMatrixDirectedGraph::hasEdge(const Node &from, const Node &to) const
{
	auto search = adj.find(from);
	if (search == adj.end()) {
		return false;
	}
	return search->second.find(to) != search->second.end();
}

std::set<DirectedEdge> AdjacencyMatrixDirectedGraph::findEdges(const Node &from, const Node &to) const
{
	std::set<DirectedEdge> out;

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == from && e.getToNode() == to) {
			out.insert(e);
		}
	}

	return out;
}

std::set<Node> AdjacencyMatrixDirectedGraph::neighbors(const Node &node) const
{
	std::set<Node> out;

	auto search = adj.find(node);
	if (search == adj.end()) {
		return out;
	}

	for (auto i : search->second) {
		if (i.second > 0) {
			out.insert(i.first);
		}
	}

	return out;
}

std::set<DirectedEdge> AdjacencyMatrixDirectedGraph::neighborEdges(const Node &node) const
{
	std::set<DirectedEdge> out;
	const std::set<Node> &nghbrs = neighbors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == node && nghbrs.find(e.getToNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyMatrixDirectedGraph::addNode(const Node &node)
{
	if (adj.find(node) != adj.end()) {
		return false;
	}

	adj[node]; // insert key
	return true;
}

bool AdjacencyMatrixDirectedGraph::removeNode(const Node &node)
{
	auto search = adj.find(node);
	if (search == adj.end()) {
		return false;
	}

	adj.erase(search);

	auto i = edges.begin();
	while (i != edges.end()) {
		if (i->getFromNode() == node) {
			edges.erase(i++);
		} else if (i->getToNode() == node) {
			adj[i->getFromNode()].erase(node);
			edges.erase(i++);
		} else {
			i++;
		}
	}

	return true;
}

bool AdjacencyMatrixDirectedGraph::addEdge(const DirectedEdge &edge)
{
	if (edges.find(edge) != edges.end()) {
		return false;
	}

	edges.insert(edge);

	adj[edge.getFromNode()][edge.getToNode()]++;
	adj[edge.getToNode()]; // insert key

	return true;
}

bool AdjacencyMatrixDirectedGraph::removeEdge(const DirectedEdge &edge)
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);

	auto &map = adj[edge.getFromNode()];
	int &val = map[edge.getToNode()];
	if (--val <= 0) {
		map.erase(edge.getToNode());
		if (map.empty()) {
			adj.erase(edge.getFromNode());
		}
	}

	return true;
}

void AdjacencyMatrixDirectedGraph::copy(IDirectedGraph *other)
{
	AdjacencyMatrixDirectedGraph *o = static_cast<AdjacencyMatrixDirectedGraph*>(other);

	edges = o->edges;
	adj = o->adj;
}

int AdjacencyMatrixDirectedGraph::compare(IDirectedGraph *other) const
{
	AdjacencyMatrixDirectedGraph *o = static_cast<AdjacencyMatrixDirectedGraph*>(other);

	auto first = std::tie(edges, adj);
	auto second = std::tie(o->edges, o->adj);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

} // namespace graph
