#include "UndirectedEdge.h"

#include <sstream>

namespace graph {

UndirectedEdge::UndirectedEdge()
	: name(label::labelFrom('u'))
{
}

UndirectedEdge::UndirectedEdge(const Node &first, const Node &second)
	: first(first)
	, second(second)
	, name(label::labelFrom('u'))
{
}

UndirectedEdge::UndirectedEdge(Node &&first, Node &&second)
	: first(std::move(first))
	, second(std::move(second))
	, name(label::labelFrom('u'))
{
}

UndirectedEdge::UndirectedEdge(const Node &first, const Node &second, const label::Label &name)
	: first(first)
	, second(second)
	, name(name)
{
}

UndirectedEdge::UndirectedEdge(Node &&first, Node &&second, label::Label &&name)
	: first(std::move(first))
	, second(std::move(second))
	, name(std::move(name))
{
}

UndirectedEdge::UndirectedEdge(const Node &first, const Node &second, int number)
	: first(first)
	, second(second)
	, name(label::labelFrom(number))
{
}

UndirectedEdge::UndirectedEdge(Node &&first, Node &&second, int number)
	: first(std::move(first))
	, second(std::move(second))
	, name(label::labelFrom(number))
{
}

UndirectedEdge::UndirectedEdge(const Node &first, const Node &second, char character)
	: first(first)
	, second(second)
	, name(label::labelFrom(character))
{
}

UndirectedEdge::UndirectedEdge(Node &&first, Node &&second, char character)
	: first(std::move(first))
	, second(std::move(second))
	, name(label::labelFrom(character))
{
}

UndirectedEdge::UndirectedEdge(const Node &first, const Node &second, const std::string &name)
	: first(first)
	, second(second)
	, name(label::labelFrom(name))
{
}

UndirectedEdge::UndirectedEdge(Node &&first, Node &&second, const std::string &name)
	: first(std::move(first))
	, second(std::move(second))
	, name(label::labelFrom(name))
{
}

UndirectedEdge::UndirectedEdge(const UndirectedEdge &other)
	: first(other.first)
	, second(other.second)
	, name(other.name)
{
}

UndirectedEdge::UndirectedEdge(UndirectedEdge &&other) noexcept
	: first(std::move(other.first))
	, second(std::move(other.second))
	, name(std::move(other.name))
{
}

UndirectedEdge &UndirectedEdge::operator=(const UndirectedEdge &other)
{
	if (this == &other) {
		return *this;
	}

	*this = UndirectedEdge(other);
	return *this;
}

UndirectedEdge &UndirectedEdge::operator=(UndirectedEdge &&other)
{
	std::swap(first, other.first);
	std::swap(second, other.second);
	std::swap(name, other.name);
	return *this;
}

GraphElement *UndirectedEdge::clone() const
{
	return new UndirectedEdge(*this);
}

GraphElement *UndirectedEdge::plunder() &&
{
	return new UndirectedEdge(std::move(*this));
}

const Node &UndirectedEdge::getFirstNode() const
{
	return first;
}

const Node &UndirectedEdge::getSecondNode() const
{
	return second;
}

const label::Label &UndirectedEdge::getName() const
{
	return name;
}

int UndirectedEdge::compare(const UndirectedEdge &other) const
{
	auto a = sortedNodes();
	auto b = other.sortedNodes();

	int res = name.getData().compare(other.name.getData());
	if (res == 0) {
		if (a.first.compare(b.second) == 0 && a.second.compare(b.first) == 0) {
			return 0;
		}
		res = a.first.compare(b.first);
		if (res == 0) {
			res = a.second.compare(b.second);
		}
	}
	return res;
}

void UndirectedEdge::operator>>(std::ostream &out) const
{
	dump(out);
}

UndirectedEdge::operator std::string() const
{
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

std::ostream &operator<<(std::ostream &out, const UndirectedEdge &node)
{
	node.dump(out);
	return out;
}

void UndirectedEdge::dump(std::ostream &os) const
{
	auto nodes = sortedNodes();
	os << "(UndirectedEdge " << nodes.first << " -> " << nodes.second << ")";
}

std::pair<Node, Node> UndirectedEdge::sortedNodes() const
{
	std::pair<Node, Node> out;

	if (first < second) {
		out.first = first;
		out.second = second;
	} else {
		out.first = second;
		out.second = first;
	}

	return out;
}

} // namespace graph

