/*
 * LabelFromXMLParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LabelFromXMLParser.h"
#include "../sax/ParserException.h"
#include "../XmlApi.hpp"
#include <algorithm>

namespace label {

Label LabelFromXMLParser::parseLabel(std::list<sax::Token>& input) const {
	return parseLabel(input, std::set<FEATURES>({FEATURES::PRIMITIVE, FEATURES::HEXAVIGESIMAL, FEATURES::OBJECT, FEATURES::LABEL_SET, FEATURES::LABEL_PAIR, FEATURES::UNIQUE_LABEL}));
}

Label LabelFromXMLParser::parseLabel(std::list<sax::Token>& input, const std::set<FEATURES>& features) const {
	if(alib::xmlApi<PrimitiveLabel>::first(input)) {
		if(!features.count(FEATURES::PRIMITIVE)) throw exception::AlibException();
		return Label(parsePrimitiveLabel(input));
	} else if(alib::xmlApi<HexavigesimalLabel>::first(input)) {
		if(!features.count(FEATURES::HEXAVIGESIMAL)) throw exception::AlibException();
		return Label(parseHexavigesimalLabel(input));
	} else if(alib::xmlApi<ObjectLabel>::first(input)) {
		if(!features.count(FEATURES::OBJECT)) throw exception::AlibException();
		return Label(parseObjectLabel(input));
	} else if(alib::xmlApi<LabelSetLabel>::first(input)) {
		if(!features.count(FEATURES::LABEL_SET)) throw exception::AlibException();
		return Label(parseLabelSetLabel(input));
	} else if(alib::xmlApi<LabelPairLabel>::first(input)) {
		if(!features.count(FEATURES::LABEL_PAIR)) throw exception::AlibException();
		return Label(parseLabelPairLabel(input));
	} else if(alib::xmlApi<UniqueLabel>::first(input)) {
		if(!features.count(FEATURES::UNIQUE_LABEL)) throw exception::AlibException();
		return Label(parseUniqueLabel(input));
	} else {
		throw sax::ParserException(sax::Token("PrimitiveLabel, HexavigesimalLabel, ObjectLabel, LabelSetLabel, LabelPairLabel, UniqueLabel", sax::Token::TokenType::START_ELEMENT), input.front());
	}
}

bool LabelFromXMLParser::first(const std::list<sax::Token>& input) const {
	if(alib::xmlApi<PrimitiveLabel>::first(input) || alib::xmlApi<ObjectLabel>::first(input) || alib::xmlApi<LabelSetLabel>::first(input) || alib::xmlApi<LabelPairLabel>::first(input)) {
		return true;
	} else {
		return false;
	}
}

PrimitiveLabel LabelFromXMLParser::parsePrimitiveLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::LABEL_PRIMITIVE_LABEL);
	PrimitiveLabel data(alib::xmlApi<primitive::Primitive>::parse(input));
	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::LABEL_PRIMITIVE_LABEL);
	return data;
}

HexavigesimalLabel LabelFromXMLParser::parseHexavigesimalLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::LABEL_HEXAVIGESIMAL_LABEL);
	HexavigesimalLabel data(std::stoi(popTokenData(input, sax::Token::TokenType::CHARACTER)));
	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::LABEL_HEXAVIGESIMAL_LABEL);
	return data;
}

ObjectLabel LabelFromXMLParser::parseObjectLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::LABEL_OBJECT_LABEL);
	ObjectLabel data(alib::xmlApi<alib::Object>::parse(input));
	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::LABEL_OBJECT_LABEL);
	return data;
}

LabelSetLabel LabelFromXMLParser::parseLabelSetLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::LABEL_LABEL_SET_LABEL);
	std::set<Label> labels;
	while(isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		labels.insert(parseLabel(input));
	}
	LabelSetLabel data(labels);
	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::LABEL_LABEL_SET_LABEL);
	return data;
}

LabelPairLabel LabelFromXMLParser::parseLabelPairLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::LABEL_LABEL_PAIR_LABEL);

	Label firstLabel = parseLabel(input);
	Label secondLabel = parseLabel(input);

	LabelPairLabel data(std::make_pair(firstLabel, secondLabel));
	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::LABEL_LABEL_PAIR_LABEL);
	return data;
}

UniqueLabel LabelFromXMLParser::parseUniqueLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, alib::Names::LABEL_UNIQUE_LABEL);

	Label firstLabel = parseLabel(input);
	primitive::Integer secondLabel = alib::xmlApi<primitive::Integer>::parse(input);

	UniqueLabel data(firstLabel, secondLabel);
	popToken(input, sax::Token::TokenType::END_ELEMENT, alib::Names::LABEL_UNIQUE_LABEL);
	return data;
}

} /* namespace label */
