/*
 * FormalRegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_H_
#define FORMAL_REG_EXP_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "FormalRegExpElement.h"
#include "../../std/visitor.hpp"
#include "../RegExpBase.h"
#include "../common/RegExpAlphabet.h"

namespace regexp {

class FormalRegExpElement;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class FormalRegExp : public std::acceptor<FormalRegExp, VisitableRegExpBase, std::acceptor<FormalRegExp, alib::VisitableObjectBase, RegExpBase> >, public RegExpAlphabet {
protected:
	FormalRegExpElement* regExp;

public:
	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	virtual RegExpBase* clone() const;

	/**
	 * @copydoc FormalRegExpElement::plunder() const
	 */
	virtual RegExpBase* plunder() &&;

	explicit FormalRegExp();
	explicit FormalRegExp(const UnboundedRegExp& other);
	explicit FormalRegExp(const std::set<alphabet::Symbol>& alphabet, const FormalRegExpElement& regExp);
	explicit FormalRegExp(std::set<alphabet::Symbol>&& alphabet, FormalRegExpElement&& regExp);
	explicit FormalRegExp(const FormalRegExpElement& regExp);
	explicit FormalRegExp(FormalRegExpElement&& regExp);

	/**
	 * Copy constructor.
	 * @param other RegExp to copy
	 */
	FormalRegExp(const FormalRegExp& other);
	FormalRegExp(FormalRegExp&& other) noexcept;
	FormalRegExp& operator =(const FormalRegExp& other);
	FormalRegExp& operator =(FormalRegExp&& other) noexcept;
	~FormalRegExp() noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const FormalRegExpElement& getRegExp() const;

	/**
	 * @return Root node of the regular expression tree
	 */
	FormalRegExpElement& getRegExp();

	/**
	 * Sets the root node of the regular expression tree. Doesn't perform copy of the regExp param,
	 * just stores it!
	 * @param regExp root node to set
	 */
	void setRegExp(const FormalRegExpElement& regExp);

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp(FormalRegExpElement&& regExp);

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet(const alphabet::Symbol & symbol);

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	virtual void operator>>(std::ostream& out) const;

	virtual int compare(const ObjectBase& other) const {
		return -other.compare(*this);
	}

	virtual int compare(const FormalRegExp& other) const;

	virtual explicit operator std::string() const;

	virtual int selfTypeId() const {
		return typeId<FormalRegExp>();
	}
};

} /* namespace regexp */

namespace std {

template<>
struct compare<regexp::FormalRegExp> {
	int operator()(const regexp::FormalRegExp& first, const regexp::FormalRegExp& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* FORMAL_REG_EXP_H_ */
