/*
 * compare.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef COMPARE_HPP_
#define COMPARE_HPP_

#include <set>
#include <list>
#include <vector>
#include <string>
#include <map>
#include <unordered_map>
#include <tuple>

namespace std {

template<class T>
struct compare {
	int operator()(T first, T second) const;
};

template<>
struct compare<bool> {
	int operator()(bool first, bool second) const {
		return first - second;
	}
};

template<>
struct compare<int> {
	int operator()(int first, int second) const {
		return first - second;
	}
};

template<>
struct compare<double> {
	int operator()(double first, double second) const {
		return first - second;
	}
};

template<class T>
struct compare<set<T>> {
	int operator()(const set<T>& first, const set<T>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		compare<T> comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); iterF++, iterS++) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

template<class T>
struct compare<list<T>> {
	int operator()(const list<T>& first, const list<T>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		compare<T> comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); iterF++, iterS++) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

template<class T>
struct compare<vector<T>> {
	int operator()(const vector<T>& first, const vector<T>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		compare<T> comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); iterF++, iterS++) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

template<class T, class R>
struct compare<pair<T, R>> {
	int operator()(const pair<T, R>& first, const pair<T, R>& second) const {
		compare<T> compT;
		compare<R> compR;

		int res = compT(first.first, second.first);
		if(res == 0) res = compR(first.second, second.second);
		return res;
	}
};

template<class T, class R>
struct compare<map<T, R>> {
	int operator()(const map<T, R>& first, const map<T, R>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		compare<std::pair<T, R>> comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); iterF++, iterS++) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

template<class T, class R>
struct compare<unordered_map<T, R>> {
	int operator()(const unordered_map<T, R>& first, const unordered_map<T, R>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		compare<R> comp;
		for(auto iter = first.begin(); iter != first.end(); iter++) {
			auto search = second.find(iter->first);
			if(search == second.end()) return -1;
			int res = comp(iter->second, search->second);
			if(res != 0) return res;
		}
		return 0;
	}
};

template<int I, typename Tuple>
struct compareTupleHelper;

template<int I, typename Tuple>
struct compareTupleHelper {
	static int compHelp(const Tuple& t1, const Tuple& t2) {
		compare<typename tuple_element<I, Tuple>::type> comp;
		int res = comp(std::get<I>(t1), std::get<I>(t2));
		if(res != 0) return res;
		return compareTupleHelper<I - 1, Tuple>::compHelp(t1, t2);
	}
};

template<class Tuple>
struct compareTupleHelper<0, Tuple> {
	static int compHelp(const Tuple& t1, const Tuple& t2) {
		compare<typename tuple_element<0, Tuple>::type> comp;
		return comp(std::get<0>(t1), std::get<0>(t2));
	}
};


template<typename ... Ts>
struct compare<tuple<Ts...>> {
	int operator()(const tuple<Ts...>& first, const tuple<Ts...>& second) const {
		return compareTupleHelper<tuple_size<std::tuple<Ts...>>::value - 1, std::tuple<Ts...>>::compHelp(first, second);
	}
};

template<>
struct compare<string> {
	int operator()(const string& first, const string& second) const {
		return first.compare(second);
	}
};


template<int I, typename Tie>
struct compareTieHelper;

template<int I, typename Tie>
struct compareTieHelper {
	static int compHelp(const Tie& t1, const Tie& t2) {
		compare<typename std::remove_const<typename std::remove_reference<typename tuple_element<I, Tie>::type>::type>::type> comp;
		int res = comp(std::get<I>(t1), std::get<I>(t2));
		if(res != 0) return res;
		return compareTieHelper<I - 1, Tie>::compHelp(t1, t2);
	}
};

template<class Tie>
struct compareTieHelper<0, Tie> {
	static int compHelp(const Tie& t1, const Tie& t2) {
		compare<typename std::remove_const<typename std::remove_reference<typename tuple_element<0, Tie>::type>::type>::type> comp;
		return comp(std::get<0>(t1), std::get<0>(t2));
	}
};


template<typename ... Ts>
struct compare<tuple<Ts&...>> {
	int operator()(const tuple<Ts&...>& first, const tuple<Ts&...>& second) const {
		return compareTieHelper<tuple_size<std::tuple<Ts&...>>::value - 1, std::tuple<Ts&...>>::compHelp(first, second);
	}
};

} /* namespace std */

#endif /* COMPARE_HPP_ */
