#include <list>
#include "RegExpTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "regexp/unbounded/UnboundedRegExp.h"

#include "factory/XmlDataFactory.hpp"
#include "factory/StringDataFactory.hpp"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))
#define CPPUNIT_EXCLUSIVE_OR(x, y) CPPUNIT_ASSERT((!(x) && (y)) || ((x) && !(y)))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( RegExpTest, "regexp" );
CPPUNIT_TEST_SUITE_REGISTRATION( RegExpTest );

void RegExpTest::setUp() {
}

void RegExpTest::tearDown() {
}

void RegExpTest::testCopyConstruct() {
	{
		regexp::UnboundedRegExp regexp;
		regexp.setAlphabet({alphabet::symbolFrom("1"), alphabet::symbolFrom("2"), alphabet::symbolFrom("3")});
		regexp::UnboundedRegExpSymbol l1 = regexp::UnboundedRegExpSymbol("1");
		regexp::UnboundedRegExpSymbol l2 = regexp::UnboundedRegExpSymbol("2");

		regexp::UnboundedRegExpConcatenation con = regexp::UnboundedRegExpConcatenation();
		con.appendElement(l1);
		con.appendElement(l2);

		regexp::UnboundedRegExpIteration ite = regexp::UnboundedRegExpIteration(l1);

		regexp::UnboundedRegExpAlternation alt = regexp::UnboundedRegExpAlternation();
		alt.appendElement(con);
		alt.appendElement(ite);

		regexp.setRegExp(alt);

		regexp::UnboundedRegExp regexp2(regexp);

		CPPUNIT_ASSERT( regexp == regexp2 );

		regexp::UnboundedRegExp regexp3(std::move(regexp));

		CPPUNIT_ASSERT( regexp2 == regexp3 );
	}
	{
		regexp::FormalRegExp regexp;
		regexp.setAlphabet({alphabet::symbolFrom("1"), alphabet::symbolFrom("2"), alphabet::symbolFrom("3")});
		regexp::FormalRegExpSymbol l1 = regexp::FormalRegExpSymbol("1");
		regexp::FormalRegExpSymbol l2 = regexp::FormalRegExpSymbol("2");

		regexp::FormalRegExpConcatenation con = regexp::FormalRegExpConcatenation(l1, l2);

		regexp::FormalRegExpIteration ite = regexp::FormalRegExpIteration(l1);

		regexp::FormalRegExpAlternation alt = regexp::FormalRegExpAlternation(con, ite);

		regexp.setRegExp(alt);

		regexp::FormalRegExp regexp2(regexp);

		CPPUNIT_ASSERT( regexp == regexp2 );

		regexp::FormalRegExp regexp3(std::move(regexp));

		CPPUNIT_ASSERT( regexp2 == regexp3 );
	}
}

void RegExpTest::testEqual() {
	{
		std::string input = "#E* #0*";
		regexp::RegExp regexp = alib::StringDataFactory::fromString<regexp::RegExp>(input);

		std::string output = alib::StringDataFactory::toString(regexp);

		CPPUNIT_ASSERT( input == output );

		regexp::RegExp regexp2 = alib::StringDataFactory::fromString<regexp::RegExp>(output);

		CPPUNIT_ASSERT( regexp == regexp2 );
	}
	{
		std::string input = "a+a";
		regexp::RegExp regexp = alib::StringDataFactory::fromString<regexp::RegExp>(input);

		std::string output = alib::StringDataFactory::toString(regexp);

		CPPUNIT_ASSERT( input == output );

		regexp::RegExp regexp2 = alib::StringDataFactory::fromString<regexp::RegExp>(output);

		CPPUNIT_ASSERT( regexp == regexp2 );
	}
	{
		std::string input = "a+a (b+a)*";
		regexp::RegExp regexp = alib::StringDataFactory::fromString<regexp::RegExp>(input);

		std::string output = alib::StringDataFactory::toString(regexp);

		CPPUNIT_ASSERT( input == output );
		std::cout << output << std::endl;

		regexp::RegExp regexp2 = alib::StringDataFactory::fromString<regexp::RegExp>(output);

		CPPUNIT_ASSERT( regexp == regexp2 );
	}
}

void RegExpTest::testXMLParser() {

	regexp::UnboundedRegExp unboundedRegexp;
	unboundedRegexp.setAlphabet({alphabet::symbolFrom("1"), alphabet::symbolFrom("2"), alphabet::symbolFrom("3")});

	regexp::UnboundedRegExpSymbol l1 = regexp::UnboundedRegExpSymbol("1");
	regexp::UnboundedRegExpSymbol l2 = regexp::UnboundedRegExpSymbol("2");

	regexp::UnboundedRegExpConcatenation con = regexp::UnboundedRegExpConcatenation();
	con.appendElement(l1);
	con.appendElement(l2);

	regexp::UnboundedRegExpIteration ite = regexp::UnboundedRegExpIteration(l1);

	regexp::UnboundedRegExpAlternation alt = regexp::UnboundedRegExpAlternation();
	alt.appendElement(con);
	alt.appendElement(ite);

	unboundedRegexp.setRegExp(alt);

	regexp::RegExp regexp(unboundedRegexp);

	{
	std::list<sax::Token> tokens = alib::XmlDataFactory::toTokens(regexp);
	std::string tmp;
	sax::SaxComposeInterface::printMemory(tmp, tokens);

	std::list<sax::Token> tokens2;
	sax::SaxParseInterface::parseMemory(tmp, tokens2);
	regexp::RegExp regexp2 = alib::XmlDataFactory::fromTokens<regexp::RegExp>(tokens2);

	CPPUNIT_ASSERT( regexp == regexp2 );
	}
	{
	std::string tmp = alib::XmlDataFactory::toString(regexp);
	regexp::RegExp regexp2 = alib::XmlDataFactory::fromString<regexp::RegExp>(tmp);

	CPPUNIT_ASSERT( regexp == regexp2 );
	}
}

void RegExpTest::testOrder() {
	regexp::UnboundedRegExpSymbol s1("1");
	regexp::UnboundedRegExpEmpty e1;
	regexp::UnboundedRegExpEpsilon e2;
	regexp::UnboundedRegExpIteration i1(s1);
	regexp::UnboundedRegExpConcatenation a1;
	a1.appendElement(s1);
	a1.appendElement(s1);
	regexp::UnboundedRegExpAlternation c1;
	c1.appendElement(s1);
	c1.appendElement(s1);

	regexp::UnboundedRegExp alt1(a1);
	regexp::UnboundedRegExp con1(c1);
	regexp::UnboundedRegExp ite1(i1);
	regexp::UnboundedRegExp emp1(e1);
	regexp::UnboundedRegExp eps1(e2);
	regexp::UnboundedRegExp sym1(s1);

	CPPUNIT_EXCLUSIVE_OR(alt1 < con1, con1 < alt1);
	CPPUNIT_EXCLUSIVE_OR(alt1 < ite1, ite1 < alt1);
	CPPUNIT_EXCLUSIVE_OR(alt1 < emp1, emp1 < alt1);
	CPPUNIT_EXCLUSIVE_OR(alt1 < eps1, eps1 < alt1);
	CPPUNIT_EXCLUSIVE_OR(alt1 < sym1, sym1 < alt1);

	CPPUNIT_EXCLUSIVE_OR(con1 < ite1, ite1 < con1);
	CPPUNIT_EXCLUSIVE_OR(con1 < emp1, emp1 < con1);
	CPPUNIT_EXCLUSIVE_OR(con1 < eps1, eps1 < con1);
	CPPUNIT_EXCLUSIVE_OR(con1 < sym1, sym1 < con1);

	CPPUNIT_EXCLUSIVE_OR(ite1 < emp1, emp1 < ite1);
	CPPUNIT_EXCLUSIVE_OR(ite1 < eps1, eps1 < ite1);
	CPPUNIT_EXCLUSIVE_OR(ite1 < sym1, sym1 < ite1);

	CPPUNIT_EXCLUSIVE_OR(emp1 < eps1, eps1 < emp1);
	CPPUNIT_EXCLUSIVE_OR(emp1 < sym1, sym1 < emp1);

	CPPUNIT_EXCLUSIVE_OR(eps1 < sym1, sym1 < eps1);

	CPPUNIT_IMPLY( alt1 < con1 && con1 < ite1, alt1 < ite1 );
	CPPUNIT_IMPLY( alt1 < con1 && con1 < emp1, alt1 < emp1 );
	CPPUNIT_IMPLY( alt1 < con1 && con1 < eps1, alt1 < eps1 );
	CPPUNIT_IMPLY( alt1 < con1 && con1 < sym1, alt1 < sym1 );

	CPPUNIT_IMPLY( alt1 < ite1 && ite1 < con1, alt1 < con1 );
	CPPUNIT_IMPLY( alt1 < ite1 && ite1 < emp1, alt1 < emp1 );
	CPPUNIT_IMPLY( alt1 < ite1 && ite1 < eps1, alt1 < eps1 );
	CPPUNIT_IMPLY( alt1 < ite1 && ite1 < sym1, alt1 < sym1 );

	CPPUNIT_IMPLY( alt1 < emp1 && emp1 < con1, alt1 < con1 );
	CPPUNIT_IMPLY( alt1 < emp1 && emp1 < ite1, alt1 < ite1 );
	CPPUNIT_IMPLY( alt1 < emp1 && emp1 < eps1, alt1 < eps1 );
	CPPUNIT_IMPLY( alt1 < emp1 && emp1 < sym1, alt1 < sym1 );

	CPPUNIT_IMPLY( alt1 < eps1 && eps1 < con1, alt1 < con1 );
	CPPUNIT_IMPLY( alt1 < eps1 && eps1 < ite1, alt1 < ite1 );
	CPPUNIT_IMPLY( alt1 < eps1 && eps1 < emp1, alt1 < emp1 );
	CPPUNIT_IMPLY( alt1 < eps1 && eps1 < sym1, alt1 < sym1 );

	CPPUNIT_IMPLY( alt1 < sym1 && sym1 < con1, alt1 < con1 );
	CPPUNIT_IMPLY( alt1 < sym1 && sym1 < ite1, alt1 < ite1 );
	CPPUNIT_IMPLY( alt1 < sym1 && sym1 < emp1, alt1 < emp1 );
	CPPUNIT_IMPLY( alt1 < sym1 && sym1 < eps1, alt1 < eps1 );



	CPPUNIT_IMPLY( con1 < alt1 && alt1 < ite1, con1 < ite1 );
	CPPUNIT_IMPLY( con1 < alt1 && alt1 < emp1, con1 < emp1 );
	CPPUNIT_IMPLY( con1 < alt1 && alt1 < eps1, con1 < eps1 );
	CPPUNIT_IMPLY( con1 < alt1 && alt1 < sym1, con1 < sym1 );

	CPPUNIT_IMPLY( con1 < ite1 && ite1 < alt1, con1 < alt1 );
	CPPUNIT_IMPLY( con1 < ite1 && ite1 < emp1, con1 < emp1 );
	CPPUNIT_IMPLY( con1 < ite1 && ite1 < eps1, con1 < eps1 );
	CPPUNIT_IMPLY( con1 < ite1 && ite1 < sym1, con1 < sym1 );

	CPPUNIT_IMPLY( con1 < emp1 && emp1 < alt1, con1 < alt1 );
	CPPUNIT_IMPLY( con1 < emp1 && emp1 < ite1, con1 < ite1 );
	CPPUNIT_IMPLY( con1 < emp1 && emp1 < eps1, con1 < eps1 );
	CPPUNIT_IMPLY( con1 < emp1 && emp1 < sym1, con1 < sym1 );

	CPPUNIT_IMPLY( con1 < eps1 && eps1 < alt1, con1 < alt1 );
	CPPUNIT_IMPLY( con1 < eps1 && eps1 < ite1, con1 < ite1 );
	CPPUNIT_IMPLY( con1 < eps1 && eps1 < emp1, con1 < emp1 );
	CPPUNIT_IMPLY( con1 < eps1 && eps1 < sym1, con1 < sym1 );

	CPPUNIT_IMPLY( con1 < sym1 && sym1 < alt1, con1 < alt1 );
	CPPUNIT_IMPLY( con1 < sym1 && sym1 < ite1, con1 < ite1 );
	CPPUNIT_IMPLY( con1 < sym1 && sym1 < emp1, con1 < emp1 );
	CPPUNIT_IMPLY( con1 < sym1 && sym1 < eps1, con1 < eps1 );



	CPPUNIT_IMPLY( ite1 < alt1 && alt1 < con1, ite1 < con1 );
	CPPUNIT_IMPLY( ite1 < alt1 && alt1 < emp1, ite1 < emp1 );
	CPPUNIT_IMPLY( ite1 < alt1 && alt1 < eps1, ite1 < eps1 );
	CPPUNIT_IMPLY( ite1 < alt1 && alt1 < sym1, ite1 < sym1 );

	CPPUNIT_IMPLY( ite1 < con1 && con1 < alt1, ite1 < alt1 );
	CPPUNIT_IMPLY( ite1 < con1 && con1 < emp1, ite1 < emp1 );
	CPPUNIT_IMPLY( ite1 < con1 && con1 < eps1, ite1 < eps1 );
	CPPUNIT_IMPLY( ite1 < con1 && con1 < sym1, ite1 < sym1 );

	CPPUNIT_IMPLY( ite1 < emp1 && emp1 < alt1, ite1 < alt1 );
	CPPUNIT_IMPLY( ite1 < emp1 && emp1 < con1, ite1 < con1 );
	CPPUNIT_IMPLY( ite1 < emp1 && emp1 < eps1, ite1 < eps1 );
	CPPUNIT_IMPLY( ite1 < emp1 && emp1 < sym1, ite1 < sym1 );

	CPPUNIT_IMPLY( ite1 < eps1 && eps1 < alt1, ite1 < alt1 );
	CPPUNIT_IMPLY( ite1 < eps1 && eps1 < con1, ite1 < con1 );
	CPPUNIT_IMPLY( ite1 < eps1 && eps1 < emp1, ite1 < emp1 );
	CPPUNIT_IMPLY( ite1 < eps1 && eps1 < sym1, ite1 < sym1 );

	CPPUNIT_IMPLY( ite1 < sym1 && sym1 < alt1, ite1 < alt1 );
	CPPUNIT_IMPLY( ite1 < sym1 && sym1 < con1, ite1 < con1 );
	CPPUNIT_IMPLY( ite1 < sym1 && sym1 < emp1, ite1 < emp1 );
	CPPUNIT_IMPLY( ite1 < sym1 && sym1 < eps1, ite1 < eps1 );



	CPPUNIT_IMPLY( emp1 < alt1 && alt1 < con1, emp1 < con1 );
	CPPUNIT_IMPLY( emp1 < alt1 && alt1 < ite1, emp1 < ite1 );
	CPPUNIT_IMPLY( emp1 < alt1 && alt1 < eps1, emp1 < eps1 );
	CPPUNIT_IMPLY( emp1 < alt1 && alt1 < sym1, emp1 < sym1 );

	CPPUNIT_IMPLY( emp1 < con1 && con1 < alt1, emp1 < alt1 );
	CPPUNIT_IMPLY( emp1 < con1 && con1 < ite1, emp1 < ite1 );
	CPPUNIT_IMPLY( emp1 < con1 && con1 < eps1, emp1 < eps1 );
	CPPUNIT_IMPLY( emp1 < con1 && con1 < sym1, emp1 < sym1 );

	CPPUNIT_IMPLY( emp1 < ite1 && ite1 < alt1, emp1 < alt1 );
	CPPUNIT_IMPLY( emp1 < ite1 && ite1 < con1, emp1 < con1 );
	CPPUNIT_IMPLY( emp1 < ite1 && ite1 < eps1, emp1 < eps1 );
	CPPUNIT_IMPLY( emp1 < ite1 && ite1 < sym1, emp1 < sym1 );

	CPPUNIT_IMPLY( emp1 < eps1 && eps1 < alt1, emp1 < alt1 );
	CPPUNIT_IMPLY( emp1 < eps1 && eps1 < con1, emp1 < con1 );
	CPPUNIT_IMPLY( emp1 < eps1 && eps1 < ite1, emp1 < ite1 );
	CPPUNIT_IMPLY( emp1 < eps1 && eps1 < sym1, emp1 < sym1 );

	CPPUNIT_IMPLY( emp1 < sym1 && sym1 < alt1, emp1 < alt1 );
	CPPUNIT_IMPLY( emp1 < sym1 && sym1 < con1, emp1 < con1 );
	CPPUNIT_IMPLY( emp1 < sym1 && sym1 < ite1, emp1 < ite1 );
	CPPUNIT_IMPLY( emp1 < sym1 && sym1 < eps1, emp1 < eps1 );



	CPPUNIT_IMPLY( eps1 < alt1 && alt1 < con1, eps1 < con1 );
	CPPUNIT_IMPLY( eps1 < alt1 && alt1 < ite1, eps1 < ite1 );
	CPPUNIT_IMPLY( eps1 < alt1 && alt1 < emp1, eps1 < emp1 );
	CPPUNIT_IMPLY( eps1 < alt1 && alt1 < sym1, eps1 < sym1 );

	CPPUNIT_IMPLY( eps1 < con1 && con1 < alt1, eps1 < alt1 );
	CPPUNIT_IMPLY( eps1 < con1 && con1 < ite1, eps1 < ite1 );
	CPPUNIT_IMPLY( eps1 < con1 && con1 < emp1, eps1 < emp1 );
	CPPUNIT_IMPLY( eps1 < con1 && con1 < sym1, eps1 < sym1 );

	CPPUNIT_IMPLY( eps1 < ite1 && ite1 < alt1, eps1 < alt1 );
	CPPUNIT_IMPLY( eps1 < ite1 && ite1 < con1, eps1 < con1 );
	CPPUNIT_IMPLY( eps1 < ite1 && ite1 < emp1, eps1 < emp1 );
	CPPUNIT_IMPLY( eps1 < ite1 && ite1 < sym1, eps1 < sym1 );

	CPPUNIT_IMPLY( eps1 < emp1 && emp1 < alt1, eps1 < alt1 );
	CPPUNIT_IMPLY( eps1 < emp1 && emp1 < con1, eps1 < con1 );
	CPPUNIT_IMPLY( eps1 < emp1 && emp1 < ite1, eps1 < ite1 );
	CPPUNIT_IMPLY( eps1 < emp1 && emp1 < sym1, eps1 < sym1 );

	CPPUNIT_IMPLY( eps1 < sym1 && sym1 < alt1, eps1 < alt1 );
	CPPUNIT_IMPLY( eps1 < sym1 && sym1 < con1, eps1 < con1 );
	CPPUNIT_IMPLY( eps1 < sym1 && sym1 < ite1, eps1 < ite1 );
	CPPUNIT_IMPLY( eps1 < sym1 && sym1 < emp1, eps1 < emp1 );



	CPPUNIT_IMPLY( sym1 < alt1 && alt1 < con1, sym1 < con1 );
	CPPUNIT_IMPLY( sym1 < alt1 && alt1 < ite1, sym1 < ite1 );
	CPPUNIT_IMPLY( sym1 < alt1 && alt1 < emp1, sym1 < emp1 );
	CPPUNIT_IMPLY( sym1 < alt1 && alt1 < eps1, sym1 < eps1 );

	CPPUNIT_IMPLY( sym1 < con1 && con1 < alt1, sym1 < alt1 );
	CPPUNIT_IMPLY( sym1 < con1 && con1 < ite1, sym1 < ite1 );
	CPPUNIT_IMPLY( sym1 < con1 && con1 < emp1, sym1 < emp1 );
	CPPUNIT_IMPLY( sym1 < con1 && con1 < eps1, sym1 < eps1 );

	CPPUNIT_IMPLY( sym1 < ite1 && ite1 < alt1, sym1 < alt1 );
	CPPUNIT_IMPLY( sym1 < ite1 && ite1 < con1, sym1 < con1 );
	CPPUNIT_IMPLY( sym1 < ite1 && ite1 < emp1, sym1 < emp1 );
	CPPUNIT_IMPLY( sym1 < ite1 && ite1 < eps1, sym1 < eps1 );

	CPPUNIT_IMPLY( sym1 < emp1 && emp1 < alt1, sym1 < alt1 );
	CPPUNIT_IMPLY( sym1 < emp1 && emp1 < con1, sym1 < con1 );
	CPPUNIT_IMPLY( sym1 < emp1 && emp1 < ite1, sym1 < ite1 );
	CPPUNIT_IMPLY( sym1 < emp1 && emp1 < eps1, sym1 < eps1 );

	CPPUNIT_IMPLY( sym1 < eps1 && eps1 < alt1, sym1 < alt1 );
	CPPUNIT_IMPLY( sym1 < eps1 && eps1 < con1, sym1 < con1 );
	CPPUNIT_IMPLY( sym1 < eps1 && eps1 < ite1, sym1 < ite1 );
	CPPUNIT_IMPLY( sym1 < eps1 && eps1 < emp1, sym1 < emp1 );
}

void RegExpTest::testOrder2() {
	regexp::UnboundedRegExpSymbol s1("1");
	regexp::UnboundedRegExpSymbol s2("2");
	regexp::UnboundedRegExpSymbol s3("3");

	regexp::UnboundedRegExpEmpty e1;
	regexp::UnboundedRegExpEpsilon e2;

	regexp::UnboundedRegExpIteration i1(s1);
	regexp::UnboundedRegExpIteration i2(s2);
	regexp::UnboundedRegExpIteration i3(s3);

	regexp::UnboundedRegExpAlternation a1;
	a1.appendElement(s1);
	a1.appendElement(s1);
	regexp::UnboundedRegExpAlternation a2;
	a2.appendElement(s1);
	a2.appendElement(s2);
	regexp::UnboundedRegExpAlternation a3;
	a3.appendElement(s1);
	a3.appendElement(s3);
	regexp::UnboundedRegExpAlternation a4;
	a4.appendElement(s2);
	a4.appendElement(s1);
	regexp::UnboundedRegExpAlternation a5;
	a5.appendElement(s2);
	a5.appendElement(s2);
	regexp::UnboundedRegExpAlternation a6;
	a6.appendElement(s2);
	a6.appendElement(s3);
	regexp::UnboundedRegExpAlternation a7;
	a7.appendElement(s3);
	a7.appendElement(s1);
	regexp::UnboundedRegExpAlternation a8;
	a8.appendElement(s3);
	a8.appendElement(s2);
	regexp::UnboundedRegExpAlternation a9;
	a9.appendElement(s3);
	a9.appendElement(s3);

	regexp::UnboundedRegExpConcatenation c1;
	c1.appendElement(s1);
	c1.appendElement(s1);
	regexp::UnboundedRegExpConcatenation c2;
	c2.appendElement(s1);
	c2.appendElement(s2);
	regexp::UnboundedRegExpConcatenation c3;
	c3.appendElement(s1);
	c3.appendElement(s3);
	regexp::UnboundedRegExpConcatenation c4;
	c4.appendElement(s2);
	c4.appendElement(s1);
	regexp::UnboundedRegExpConcatenation c5;
	c5.appendElement(s2);
	c5.appendElement(s2);
	regexp::UnboundedRegExpConcatenation c6;
	c6.appendElement(s2);
	c6.appendElement(s3);
	regexp::UnboundedRegExpConcatenation c7;
	c7.appendElement(s3);
	c7.appendElement(s1);
	regexp::UnboundedRegExpConcatenation c8;
	c8.appendElement(s3);
	c8.appendElement(s2);
	regexp::UnboundedRegExpConcatenation c9;
	c9.appendElement(s3);
	c9.appendElement(s3);

	regexp::UnboundedRegExp alt1(a1);
	regexp::UnboundedRegExp alt2(a2);
	regexp::UnboundedRegExp alt3(a3);
	regexp::UnboundedRegExp alt4(a4);
	regexp::UnboundedRegExp alt5(a5);
	regexp::UnboundedRegExp alt6(a6);
	regexp::UnboundedRegExp alt7(a7);
	regexp::UnboundedRegExp alt8(a8);
	regexp::UnboundedRegExp alt9(a9);

	regexp::UnboundedRegExp con1(c1);
	regexp::UnboundedRegExp con2(c2);
	regexp::UnboundedRegExp con3(c3);
	regexp::UnboundedRegExp con4(c4);
	regexp::UnboundedRegExp con5(c5);
	regexp::UnboundedRegExp con6(c6);
	regexp::UnboundedRegExp con7(c7);
	regexp::UnboundedRegExp con8(c8);
	regexp::UnboundedRegExp con9(c9);

	regexp::UnboundedRegExp ite1(i1);
	regexp::UnboundedRegExp ite2(i2);
	regexp::UnboundedRegExp ite3(i3);

	regexp::UnboundedRegExp emp1(e1);
	regexp::UnboundedRegExp eps1(e2);

	regexp::UnboundedRegExp sym1(s1);
	regexp::UnboundedRegExp sym2(s2);
	regexp::UnboundedRegExp sym3(s3);



	CPPUNIT_ASSERT(alt1 == alt1);
	CPPUNIT_ASSERT(alt1 < alt2);
	CPPUNIT_ASSERT(alt1 < alt3);
	CPPUNIT_ASSERT(alt1 < alt4);
	CPPUNIT_ASSERT(alt1 < alt5);
	CPPUNIT_ASSERT(alt1 < alt6);
	CPPUNIT_ASSERT(alt1 < alt7);
	CPPUNIT_ASSERT(alt1 < alt8);
	CPPUNIT_ASSERT(alt1 < alt9);

	CPPUNIT_ASSERT(alt2 > alt1);
	CPPUNIT_ASSERT(alt2 == alt2);
	CPPUNIT_ASSERT(alt2 < alt3);
	CPPUNIT_ASSERT(alt2 < alt4);
	CPPUNIT_ASSERT(alt2 < alt5);
	CPPUNIT_ASSERT(alt2 < alt6);
	CPPUNIT_ASSERT(alt2 < alt7);
	CPPUNIT_ASSERT(alt2 < alt8);
	CPPUNIT_ASSERT(alt2 < alt9);

	CPPUNIT_ASSERT(alt3 > alt1);
	CPPUNIT_ASSERT(alt3 > alt2);
	CPPUNIT_ASSERT(alt3 == alt3);
	CPPUNIT_ASSERT(alt3 < alt4);
	CPPUNIT_ASSERT(alt3 < alt5);
	CPPUNIT_ASSERT(alt3 < alt6);
	CPPUNIT_ASSERT(alt3 < alt7);
	CPPUNIT_ASSERT(alt3 < alt8);
	CPPUNIT_ASSERT(alt3 < alt9);

	CPPUNIT_ASSERT(alt4 > alt1);
	CPPUNIT_ASSERT(alt4 > alt2);
	CPPUNIT_ASSERT(alt4 > alt3);
	CPPUNIT_ASSERT(alt4 == alt4);
	CPPUNIT_ASSERT(alt4 < alt5);
	CPPUNIT_ASSERT(alt4 < alt6);
	CPPUNIT_ASSERT(alt4 < alt7);
	CPPUNIT_ASSERT(alt4 < alt8);
	CPPUNIT_ASSERT(alt4 < alt9);

	CPPUNIT_ASSERT(alt5 > alt1);
	CPPUNIT_ASSERT(alt5 > alt2);
	CPPUNIT_ASSERT(alt5 > alt3);
	CPPUNIT_ASSERT(alt5 > alt4);
	CPPUNIT_ASSERT(alt5 == alt5);
	CPPUNIT_ASSERT(alt5 < alt6);
	CPPUNIT_ASSERT(alt5 < alt7);
	CPPUNIT_ASSERT(alt5 < alt8);
	CPPUNIT_ASSERT(alt5 < alt9);

	CPPUNIT_ASSERT(alt6 > alt1);
	CPPUNIT_ASSERT(alt6 > alt2);
	CPPUNIT_ASSERT(alt6 > alt3);
	CPPUNIT_ASSERT(alt6 > alt4);
	CPPUNIT_ASSERT(alt6 > alt5);
	CPPUNIT_ASSERT(alt6 == alt6);
	CPPUNIT_ASSERT(alt6 < alt7);
	CPPUNIT_ASSERT(alt6 < alt8);
	CPPUNIT_ASSERT(alt6 < alt9);

	CPPUNIT_ASSERT(alt7 > alt1);
	CPPUNIT_ASSERT(alt7 > alt2);
	CPPUNIT_ASSERT(alt7 > alt3);
	CPPUNIT_ASSERT(alt7 > alt4);
	CPPUNIT_ASSERT(alt7 > alt5);
	CPPUNIT_ASSERT(alt7 > alt6);
	CPPUNIT_ASSERT(alt7 == alt7);
	CPPUNIT_ASSERT(alt7 < alt8);
	CPPUNIT_ASSERT(alt7 < alt9);

	CPPUNIT_ASSERT(alt8 > alt1);
	CPPUNIT_ASSERT(alt8 > alt2);
	CPPUNIT_ASSERT(alt8 > alt3);
	CPPUNIT_ASSERT(alt8 > alt4);
	CPPUNIT_ASSERT(alt8 > alt5);
	CPPUNIT_ASSERT(alt8 > alt6);
	CPPUNIT_ASSERT(alt8 > alt7);
	CPPUNIT_ASSERT(alt8 == alt8);
	CPPUNIT_ASSERT(alt8 < alt9);

	CPPUNIT_ASSERT(alt9 > alt1);
	CPPUNIT_ASSERT(alt9 > alt2);
	CPPUNIT_ASSERT(alt9 > alt3);
	CPPUNIT_ASSERT(alt9 > alt4);
	CPPUNIT_ASSERT(alt9 > alt5);
	CPPUNIT_ASSERT(alt9 > alt6);
	CPPUNIT_ASSERT(alt9 > alt7);
	CPPUNIT_ASSERT(alt9 > alt8);
	CPPUNIT_ASSERT(alt9 == alt9);



	CPPUNIT_ASSERT(con1 == con1);
	CPPUNIT_ASSERT(con1 < con2);
	CPPUNIT_ASSERT(con1 < con3);
	CPPUNIT_ASSERT(con1 < con4);
	CPPUNIT_ASSERT(con1 < con5);
	CPPUNIT_ASSERT(con1 < con6);
	CPPUNIT_ASSERT(con1 < con7);
	CPPUNIT_ASSERT(con1 < con8);
	CPPUNIT_ASSERT(con1 < con9);

	CPPUNIT_ASSERT(con2 > con1);
	CPPUNIT_ASSERT(con2 == con2);
	CPPUNIT_ASSERT(con2 < con3);
	CPPUNIT_ASSERT(con2 < con4);
	CPPUNIT_ASSERT(con2 < con5);
	CPPUNIT_ASSERT(con2 < con6);
	CPPUNIT_ASSERT(con2 < con7);
	CPPUNIT_ASSERT(con2 < con8);
	CPPUNIT_ASSERT(con2 < con9);

	CPPUNIT_ASSERT(con3 > con1);
	CPPUNIT_ASSERT(con3 > con2);
	CPPUNIT_ASSERT(con3 == con3);
	CPPUNIT_ASSERT(con3 < con4);
	CPPUNIT_ASSERT(con3 < con5);
	CPPUNIT_ASSERT(con3 < con6);
	CPPUNIT_ASSERT(con3 < con7);
	CPPUNIT_ASSERT(con3 < con8);
	CPPUNIT_ASSERT(con3 < con9);

	CPPUNIT_ASSERT(con4 > con1);
	CPPUNIT_ASSERT(con4 > con2);
	CPPUNIT_ASSERT(con4 > con3);
	CPPUNIT_ASSERT(con4 == con4);
	CPPUNIT_ASSERT(con4 < con5);
	CPPUNIT_ASSERT(con4 < con6);
	CPPUNIT_ASSERT(con4 < con7);
	CPPUNIT_ASSERT(con4 < con8);
	CPPUNIT_ASSERT(con4 < con9);

	CPPUNIT_ASSERT(con5 > con1);
	CPPUNIT_ASSERT(con5 > con2);
	CPPUNIT_ASSERT(con5 > con3);
	CPPUNIT_ASSERT(con5 > con4);
	CPPUNIT_ASSERT(con5 == con5);
	CPPUNIT_ASSERT(con5 < con6);
	CPPUNIT_ASSERT(con5 < con7);
	CPPUNIT_ASSERT(con5 < con8);
	CPPUNIT_ASSERT(con5 < con9);

	CPPUNIT_ASSERT(con6 > con1);
	CPPUNIT_ASSERT(con6 > con2);
	CPPUNIT_ASSERT(con6 > con3);
	CPPUNIT_ASSERT(con6 > con4);
	CPPUNIT_ASSERT(con6 > con5);
	CPPUNIT_ASSERT(con6 == con6);
	CPPUNIT_ASSERT(con6 < con7);
	CPPUNIT_ASSERT(con6 < con8);
	CPPUNIT_ASSERT(con6 < con9);

	CPPUNIT_ASSERT(con7 > con1);
	CPPUNIT_ASSERT(con7 > con2);
	CPPUNIT_ASSERT(con7 > con3);
	CPPUNIT_ASSERT(con7 > con4);
	CPPUNIT_ASSERT(con7 > con5);
	CPPUNIT_ASSERT(con7 > con6);
	CPPUNIT_ASSERT(con7 == con7);
	CPPUNIT_ASSERT(con7 < con8);
	CPPUNIT_ASSERT(con7 < con9);

	CPPUNIT_ASSERT(con8 > con1);
	CPPUNIT_ASSERT(con8 > con2);
	CPPUNIT_ASSERT(con8 > con3);
	CPPUNIT_ASSERT(con8 > con4);
	CPPUNIT_ASSERT(con8 > con5);
	CPPUNIT_ASSERT(con8 > con6);
	CPPUNIT_ASSERT(con8 > con7);
	CPPUNIT_ASSERT(con8 == con8);
	CPPUNIT_ASSERT(con8 < con9);

	CPPUNIT_ASSERT(con9 > con1);
	CPPUNIT_ASSERT(con9 > con2);
	CPPUNIT_ASSERT(con9 > con3);
	CPPUNIT_ASSERT(con9 > con4);
	CPPUNIT_ASSERT(con9 > con5);
	CPPUNIT_ASSERT(con9 > con6);
	CPPUNIT_ASSERT(con9 > con7);
	CPPUNIT_ASSERT(con9 > con8);
	CPPUNIT_ASSERT(con9 == con9);


	CPPUNIT_ASSERT(ite1 == ite1);
	CPPUNIT_ASSERT(ite1 < ite2);
	CPPUNIT_ASSERT(ite1 < ite3);

	CPPUNIT_ASSERT(ite2 > ite1);
	CPPUNIT_ASSERT(ite2 == ite2);
	CPPUNIT_ASSERT(ite2 < ite3);

	CPPUNIT_ASSERT(ite3 > ite1);
	CPPUNIT_ASSERT(ite3 > ite2);
	CPPUNIT_ASSERT(ite3 == ite3);



	CPPUNIT_ASSERT(emp1 == emp1);



	CPPUNIT_ASSERT(eps1 == eps1);



	CPPUNIT_ASSERT(sym1 == sym1);
	CPPUNIT_ASSERT(sym1 < sym2);
	CPPUNIT_ASSERT(sym1 < sym3);

	CPPUNIT_ASSERT(sym2 > sym1);
	CPPUNIT_ASSERT(sym2 == sym2);
	CPPUNIT_ASSERT(sym2 < sym3);

	CPPUNIT_ASSERT(sym3 > sym1);
	CPPUNIT_ASSERT(sym3 > sym2);
	CPPUNIT_ASSERT(sym3 == sym3);
}

