/*
 * arand.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Tomas Pecka
 */

#include <tclap/CmdLine.h>

#include <automaton/FSM/NFA.h>
#include <regexp/unbounded/UnboundedRegExp.h>
#include <factory/XmlDataFactory.hpp>
#include "automaton/generate/RandomAutomatonFactory.h"
#include "grammar/generate/RandomGrammarFactory.h"
#include "regexp/generate/RandomRegExpFactory.h"
#include "string/generate/RandomStringFactory.h"

int main(int argc, char* argv[]) {
	try {
		TCLAP::CmdLine cmd("Random generator binary", ' ', "0.01");

		std::vector<std::string> allowed;
		allowed.push_back("FSM");
		allowed.push_back("CFG");
		allowed.push_back("RE");
		allowed.push_back("ST");
		TCLAP::ValuesConstraint<std::string> allowedVals( allowed );

		TCLAP::ValueArg<std::string> type(	"t",	"type",		"Type of generated structure",			true,	"FSM",	&allowedVals);
		cmd.add( type );

		TCLAP::ValueArg<int> alphabetSize(	"",	"terminals",	"Number of terminals/alphabet size",		false,	3,	"integer");
		cmd.add( alphabetSize );

		TCLAP::ValueArg<double> density(	"",	"density",	"Automaton's/grammar's transition density",	false,	10,	"double");
		cmd.add( density );

		TCLAP::ValueArg<int> nonterminals(	"",	"nonterminals",	"Number of grammar's nononterminals",		false,	5,	"integer");
		cmd.add( nonterminals );

		TCLAP::ValueArg<int> states(		"",	"states",	"Number of automaton's states",			false,	5,	"integer");
		cmd.add( states );

		TCLAP::ValueArg<int> terminalNodes(	"",	"leaves",	"Number of tree's/regexp's terminal nodes",	false,	5,	"integer");
		cmd.add( terminalNodes );

		TCLAP::ValueArg<int> height(		"",	"height",	"Height of the tree/regexp",			false,	5,	"integer");
		cmd.add( height );

		TCLAP::ValueArg<int> length(		"",	"length",	"Length of the string",				false,	5,	"integer");
		cmd.add( length );

		cmd.parse(argc,argv);

		if(!type.isSet()) throw exception::AlibException("Type is not defined.");
		if( type.getValue() == "FSM" ) {

			if(!density.isSet()) throw exception::AlibException("Density is not defined.");
			if( density.getValue() < 1 || density.getValue() > 100 )
			{
				//TODO: floating point arithmetic
				throw exception::AlibException("You must specify density as a number between 1 and 100.");
			}

			if(!states.isSet()) throw exception::AlibException("Number of states is not defined.");
			if(!alphabetSize.isSet()) throw exception::AlibException("Alphabet size is not defined.");

			automaton::NFA res = automaton::generate::RandomAutomatonFactory::generateNFA( states.getValue(), alphabetSize.getValue(), density.getValue() );
			alib::XmlDataFactory::toStdout(res);
		} else if( type.getValue() == "CFG" ) {

			if(!density.isSet()) throw exception::AlibException("Density is not defined.");
			if( density.getValue() < 1 || density.getValue() > 100 )
			{
				//TODO: floating point arithmetic
				throw exception::AlibException("You must specify density as a number between 1 and 100.");
			}

			if(!nonterminals.isSet()) throw exception::AlibException("Number of nonterminals is not defined.");
			if(!alphabetSize.isSet()) throw exception::AlibException("Alphabet size is not defined.");

			grammar::CFG res = grammar::generate::RandomGrammarFactory::generateCFG( nonterminals.getValue(), alphabetSize.getValue(), density.getValue() );
			alib::XmlDataFactory::toStdout(res);
		} else if( type.getValue() == "RE" ) {

			if(!height.isSet()) throw exception::AlibException("Height is not defined.");

			if(!terminalNodes.isSet()) throw exception::AlibException("Number of terminal nodes is not defined.");
			if(!alphabetSize.isSet()) throw exception::AlibException("Alphabet size is not defined.");

			regexp::UnboundedRegExp res = regexp::generate::RandomRegExpFactory::generateUnboundedRegExp(terminalNodes.getValue(), height.getValue(), alphabetSize.getValue() );
			alib::XmlDataFactory::toStdout(res);
		} else if( type.getValue() == "ST" ) {

			if(!length.isSet()) throw exception::AlibException("Size is not defined.");

			if(!alphabetSize.isSet()) throw exception::AlibException("Alphabet size is not defined.");

			string::LinearString res = string::generate::RandomStringFactory::generateLinearString(length.getValue(), alphabetSize.getValue() );
			alib::XmlDataFactory::toStdout(res);
		} else {
			throw exception::AlibException("Invalid type.");
		}

		return 0;
	} catch( const exception::AlibException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
