/*
 * astat.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <vector>

#include <factory/XmlDataFactory.hpp>
#include <exception/AlibException.h>

#include "AutomataSettings.h"
#include "AutomataStat.h"

#include "SetSettings.h"
#include "SetStat.h"

PrintingOptions translatePrintingOptions(std::string value) {
	if(value == "noop") {
		return PrintingOptions::NOOP;
	} else if(value == "print") {
		return PrintingOptions::PRINT;
	} else if(value == "quantity") {
		return PrintingOptions::QUANTITY;
	} else if(value == "both") {
		return PrintingOptions::BOTH;
	} else {
		throw exception::AlibException("Invalid Argument in translate printing options (" + value + ")");
	}
}

int main(int argc, char* argv[]) {
	try {
		TCLAP::CmdLine cmd("Prints usefull information about automata, grammars, regexps, ...", ' ', "0.01");

		std::vector<std::string> allowedPrintingOptions {"noop", "print", "quantity", "both"};
		TCLAP::ValuesConstraint<std::string> allowedPrintingOptionsVals( allowedPrintingOptions );

		std::vector<std::string> automataPrintingOptions {"states", "finalStates", "initialStates", "inputAlphabet", "transitions"};
		TCLAP::ValuesConstraint<std::string> automataPrintingOptionsVals( automataPrintingOptions );

		TCLAP::ValueArg<std::string> presentation(	"p",	"presentation",	"How to print data",		true,	"noop",	&allowedPrintingOptionsVals );
		cmd.add(presentation);

		TCLAP::MultiArg<std::string> automaton(		"a",	"automaton",	"Display stats of automata",	false,	&automataPrintingOptionsVals);

		TCLAP::SwitchArg setContainer(			"s",	"set",		"Display stats of sets",	false);

		std::vector<TCLAP::Arg*> xorlist;
		xorlist.push_back(&automaton);
		xorlist.push_back(&setContainer);
		cmd.xorAdd( xorlist );

		TCLAP::ValueArg<std::string> file(		"i",	"input",	"Read from file",		false,	"-",	"file");
		cmd.add( file );

		cmd.parse(argc, argv);

		PrintingOptions printingOption = translatePrintingOptions(presentation.getValue());

		std::list<sax::Token> tokens;
		if(file.isSet()) {
			if(file.getValue() == "-") {
				sax::SaxParseInterface::parseStdin(tokens);
			} else {
				sax::SaxParseInterface::parseFile(file.getValue(), tokens);
			}
		} else {
			sax::SaxParseInterface::parseStdin(tokens);
		}

		if( alib::XmlDataFactory::first<automaton::Automaton>(tokens) && automaton.isSet()) {
			AutomataSettings settings = {PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP, PrintingOptions::NOOP };
			std::set<std::string> values {automaton.getValue().begin(), automaton.getValue().end()};

			if(values.count("states")) {
				settings.states = printingOption;
			}
			if(values.count("finalStates")) {
				settings.finalStates = printingOption;
			}
			if(values.count("initialStates")) {
				settings.initialStates = printingOption;
			}
			if(values.count("inputAlphabet")) {
				settings.inputAlphabet = printingOption;
			}
			if(values.count("transitions")) {
				settings.transitions = printingOption;
			}

			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(tokens);
			AutomataStat::stat(automaton, settings);

			return 0;
		} else if( alib::XmlDataFactory::first<grammar::Grammar>(tokens) /* && grammar.isSet() */) {
			grammar::Grammar grammar = alib::XmlDataFactory::fromTokens<grammar::Grammar>(tokens);

			return 0;
		} else if( alib::XmlDataFactory::first<regexp::RegExp>(tokens) /* && regexp.isSet() */) {
			regexp::RegExp regexp = alib::XmlDataFactory::fromTokens<regexp::RegExp>(tokens);

			return 0;
		} else if( alib::XmlDataFactory::first<container::ObjectsSet>(tokens) && setContainer.isSet()) {
			SetSettings settings = {PrintingOptions::NOOP};
			settings.general = printingOption;

			container::ObjectsSet data = alib::XmlDataFactory::fromTokens<container::ObjectsSet>(tokens);
			SetStat::stat(data, settings);

			return 0;
		} else {
			throw exception::AlibException( "Input not recognized." );
			return 1;
		}
	} catch ( const exception::AlibException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch (const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch (...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
