/*
 * atrim.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Tomas Pecka
 */

#include <tclap/CmdLine.h>

#include <factory/XmlDataFactory.hpp>
#include <exception/AlibException.h>

#include "grammar/simplify/Trim.h"
#include "grammar/simplify/UnproductiveSymbolsRemover.h"
#include "grammar/simplify/UnreachableSymbolsRemover.h"
#include "automaton/simplify/Trim.h"
#include "automaton/simplify/UselessStatesRemover.h"
#include "automaton/simplify/UnreachableStatesRemover.h"
#include "regexp/simplify/RegExpOptimize.h"

grammar::Grammar trimGrammar(const grammar::Grammar& g, bool del_unreachable, bool del_unproductive) {
	if( del_unreachable && del_unproductive )
		return grammar::simplify::Trim::trim( g );
	if( del_unreachable )
		return grammar::simplify::UnreachableSymbolsRemover::remove( g );
	if( del_unproductive )
		return grammar::simplify::UnproductiveSymbolsRemover::remove( g );
	return grammar::simplify::Trim::trim( g );
}

automaton::Automaton trimAutomaton(const automaton::Automaton& g, bool del_unreachable, bool del_useless) {
	if( del_unreachable && del_useless )
		return automaton::simplify::Trim::trim( g );
	if( del_unreachable )
		return automaton::simplify::UnreachableStatesRemover::remove( g );
	if( del_useless )
		return automaton::simplify::UselessStatesRemover::remove( g );
	return automaton::simplify::Trim::trim( g );
}

regexp::RegExp optimizeRegExp(const regexp::RegExp& r) {
	return regexp::simplify::RegExpOptimize::optimize( r );
}

int main(int argc, char* argv[]) {
	try {
		TCLAP::CmdLine cmd("Removes unreachable and useless states from FSM, productive and unreachable nonterminals from CFG. Simplifies representation of RE", ' ', "0.01");

		TCLAP::SwitchArg useless(		"u",	"useless",	"Removes useless states (or symbols). (works with FSM or CFG)"		);
		cmd.add( useless );

		TCLAP::SwitchArg unreachable(		"r",	"unreachable",	"Removes unreachable states (or symbols). (works with FSM or CFG)"	);
		cmd.add( unreachable );

		TCLAP::SwitchArg simplify(		"s",	"simplify",	"Simplifies representation. (works with RE only)"			);
		cmd.add( simplify );

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Input to trim",							false,	"-",	"file");
		cmd.add( input );

		cmd.parse(argc, argv);

		std::list<sax::Token> tokens;
		if(input.isSet()) {
			if(input.getValue() == "-") {
				sax::SaxParseInterface::parseStdin(tokens);
			} else {
				sax::SaxParseInterface::parseFile(input.getValue(), tokens);
			}
		} else {
			sax::SaxParseInterface::parseStdin(tokens);
		}

		if( alib::XmlDataFactory::first<automaton::Automaton>(tokens)) {
			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(tokens);

			automaton::Automaton res = trimAutomaton(automaton, unreachable.getValue(), useless.getValue() );

			alib::XmlDataFactory::toStdout( res );
			return 0;
		} else if( alib::XmlDataFactory::first<grammar::Grammar>(tokens)) {
			grammar::Grammar grammar = alib::XmlDataFactory::fromTokens<grammar::Grammar>(tokens);

			grammar::Grammar res = trimGrammar(grammar, unreachable.getValue(), useless.getValue() );

			alib::XmlDataFactory::toStdout( res );
			return 0;
		} else if( alib::XmlDataFactory::first<regexp::RegExp>(tokens)) {
			regexp::RegExp regexp = alib::XmlDataFactory::fromTokens<regexp::RegExp>(tokens);

			regexp::RegExp res = optimizeRegExp(regexp );

			alib::XmlDataFactory::toStdout( res );
			return 0;
		} else {
			throw exception::AlibException( "Input not recognized." );
			return 1;
		}
	} catch( const exception::AlibException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
